import csv

import numpy as np

from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Optional
from collections import Counter
from dataclasses import dataclass
from cftool.misc import print_info
from cftool.misc import print_warning
from cftool.misc import shallow_copy_dict
from cftool.misc import ISerializable
from cftool.misc import PureFromInfoMixin
from cftool.array import is_float

from ....types import data_type
from ....schema import DataTypes
from ....schema import DataBundle
from ....schema import INoInitDataBlock


TArrayPair = Tuple[np.ndarray, Optional[np.ndarray]]


class Converter(PureFromInfoMixin, ISerializable):
    column_name: str
    is_label: bool
    dtype: DataTypes
    mapping: Dict[str, int]
    counter: Dict[str, int]
    default_value: Optional[int]
    has_nan: bool
    has_empty: bool

    def to_info(self) -> Dict[str, Any]:
        return dict(
            column_name=self.column_name,
            is_label=self.is_label,
            dtype=self.dtype,
            mapping=self.mapping,
            counter=self.counter,
            default_value=self.default_value,
            has_nan=self.has_nan,
            has_empty=self.has_empty,
        )

    def _fit_with_custom_mapping(
        self,
        line: List[str],
        column_name: str,
        custom_mapping: Dict[str, int],
    ) -> None:
        self.mapping = shallow_copy_dict(custom_mapping)
        counter: Counter = Counter()
        for e in line:
            if e.lower() == "nan":
                e = "nan"
            if e in self.mapping:
                counter[e] += 1
            else:
                if self.default_value is None:
                    raise ValueError(
                        f"[{column_name}] OOD samples detected and "
                        "`custom_mapping` is provided, please specify "
                        "the `default_value` as well"
                    )
                counter["default"] += 1
        self.counter = dict(counter)

    @classmethod
    def fit(
        cls,
        column_name: str,
        line: List[str],
        is_label: bool,
        custom_mapping: Optional[Dict[str, int]],
        default_value: Optional[int],
    ) -> "Converter":
        self: Converter = cls()
        self.column_name = column_name
        self.is_label = is_label
        not_empty = [e for e in line if e]
        self.has_empty = len(not_empty) != len(line)
        self.default_value = default_value
        try:
            array = np.array(not_empty, np.float64)
        except ValueError:
            self.has_nan = False
            self.dtype = DataTypes.STRING
            if custom_mapping is not None:
                self._fit_with_custom_mapping(line, column_name, custom_mapping)
            else:
                counter = Counter(line)
                filtered_counter: Counter = Counter()
                for e, count in counter.most_common():
                    if e.lower() == "nan":
                        e = "nan"
                        self.has_nan = True
                    filtered_counter[e] += count
                most_common = filtered_counter.most_common()
                # smaller index -> more frequent
                self.mapping = {k: i for i, (k, _) in enumerate(most_common)}
                self.counter = dict(filtered_counter)
            return self
        nan_mask = np.isnan(array)
        not_nan_mask = ~nan_mask
        not_nan = array[not_nan_mask]
        self.has_nan = len(array) != len(not_nan)
        int_array = not_nan.astype(int)
        int_float_array = int_array.astype(np.float64)
        if not np.allclose(not_nan, int_float_array):
            self.mapping = {}
            self.counter = {}
            self.dtype = DataTypes.FLOAT
        else:
            self.dtype = DataTypes.INT
            if custom_mapping is not None:
                self._fit_with_custom_mapping(line, column_name, custom_mapping)
            else:
                if self.has_nan:
                    int_array = int_array[not_nan_mask]
                unique, counts = np.unique(int_array, return_counts=True)
                self.mapping = {}
                self.counter = {}
                unique_list = unique.tolist()
                counts_list = counts.tolist()
                for i, idx in enumerate(np.argsort(counts).tolist()[::-1]):
                    str_original = str(unique_list[idx])
                    self.mapping[str_original] = i
                    self.counter[str_original] = counts_list[idx]
        return self

    @classmethod
    def get_placeholder(cls, column_name: str, is_label: bool) -> "Converter":
        self: Converter = cls()
        self.column_name = column_name
        self.is_label = is_label
        self.dtype = DataTypes.FLOAT
        self.mapping = {}
        self.counter = {}
        self.default_value = None
        self.has_nan = False
        self.has_empty = False
        return self

    def convert(self, line: List[str], *, verbose: bool) -> np.ndarray:
        if self.dtype == DataTypes.FLOAT:
            return np.array([e if e else "nan" for e in line], np.float64)
        ood = 0
        method = None
        converted = []
        for e in line:
            if e.lower() == "nan":
                e = "nan"
            elif self.dtype == DataTypes.INT:
                e = str(int(e))
            mapped = self.mapping.get(e)
            if mapped is None:
                ood += 1
                if self.default_value is not None:
                    method = f"replaced with default value ({self.default_value})"
                    mapped = self.default_value
                elif self.has_nan:
                    method = "replaced with nan"
                    mapped = self.mapping["nan"]
                elif self.has_empty:
                    method = "replaced with empty string"
                    mapped = self.mapping[""]
                else:
                    method = "replaced with most frequent"
                    mapped = 0
            converted.append(mapped)
        if ood > 0:
            ratio = ood / len(converted)
            if verbose:
                print_warning(
                    f"[{self.column_name}] OOD samples detected "
                    f"({ood}/{len(converted)}={ratio:8.6f}), {method}"
                )
        return np.array(converted, np.float64)

    def __str__(self) -> str:
        postfix = (
            "" if self.dtype == DataTypes.STRING else f", has_empty={self.has_empty}"
        )
        return f"Converter({self.dtype}{postfix})"

    __repr__ = __str__


@dataclass
class MLFileProcessorConfig:
    delimiter: str = ","
    has_header: bool = True
    label_names: Optional[List[str]] = None
    label_indices: Optional[List[int]] = None
    contain_labels: bool = True
    auto_convert_labels: bool = True
    custom_mappings: Optional[Dict[str, Dict[str, int]]] = None
    default_values: Optional[Dict[str, int]] = None


get_x_column_name = lambda idx: f"f{idx}"
get_y_column_name = lambda idx: f"l{idx}"


@INoInitDataBlock.register("ml_file_parser")
class FileParserBlock(INoInitDataBlock):
    config: MLFileProcessorConfig  # type: ignore
    delimiter: str
    has_header: bool
    contain_labels: bool
    auto_convert_labels: bool
    custom_mappings: Optional[Dict[str, Dict[str, int]]]
    default_values: Optional[Dict[str, int]] = None
    feature_header: List[str]
    label_header: List[str]
    all_header: List[str]
    converters: Dict[str, Converter]
    is_placeholder: bool

    # inheritance

    def to_info(self) -> Dict[str, Any]:
        return dict(
            delimiter=self.delimiter,
            has_header=self.has_header,
            contain_labels=self.contain_labels,
            auto_convert_labels=self.auto_convert_labels,
            custom_mappings=self.custom_mappings,
            default_values=self.default_values,
            feature_header=self.feature_header,
            label_header=self.label_header,
            all_header=self.all_header,
            converters={k: v.to_info() for k, v in self.converters.items()},
            is_placeholder=self.is_placeholder,
        )

    def from_info(self, info: Dict[str, Any]) -> None:
        super().from_info(info)
        for k, v in self.converters.items():
            converter = Converter()
            converter.from_info(v)
            self.converters[k] = converter

    def transform(self, bundle: DataBundle, for_inference: bool) -> DataBundle:
        return self._transform(bundle, for_inference)

    def fit_transform(self, bundle: DataBundle) -> DataBundle:
        self.delimiter = self.config.delimiter
        self.has_header = self.config.has_header
        self.contain_labels = self.config.contain_labels
        self.auto_convert_labels = self.config.auto_convert_labels
        self.custom_mappings = self.config.custom_mappings or {}
        self.default_values = self.config.default_values or {}
        if not self._check_file(bundle.x_train):
            self.is_placeholder = True
            self._setup_none_file(bundle.x_train, bundle.y_train)
            return bundle
        self.is_placeholder = False
        # sanity check
        if bundle.y_train is not None:
            raise ValueError("`y_train` should not be provided for `FileParserBlock`")
        if bundle.y_valid is not None:
            raise ValueError("`y_valid` should not be provided for `FileParserBlock`")
        # read raw train data
        if not isinstance(bundle.x_train, str):
            raise ValueError("`FileParserBlock` should be used on file inputs")
        all_header, data = self._read(bundle.x_train)
        num_columns = len(data[0])
        # read raw valid data
        if bundle.x_valid is None:
            valid_data = None
        else:
            if not isinstance(bundle.x_valid, str):
                raise ValueError("`FileParserBlock` should be used on file inputs")
            all_valid_header, valid_data = self._read(bundle.x_valid)
            if self.config.has_header and all_header != all_valid_header:
                raise ValueError(
                    "train header does not match valid header:\n"  # type: ignore
                    f"> train header | {', '.join(all_header)}\n"  # type: ignore
                    f"> valid header | {', '.join(all_valid_header)}\n"  # type: ignore
                )
        # handle label columns & header
        if not self.config.contain_labels:
            msg = "`contain_labels` should be True in `FileParserBlock.fit_transform`"
            raise ValueError(msg)
        label_names = self.config.label_names
        label_indices = self.config.label_indices
        if not label_names and not label_indices:
            msg = "neither `label_names` nor `label_indices` is provided, `[-1]` will be used"
            print_warning(msg)
            label_indices = self.config.label_indices = [-1]
        if label_indices is not None:
            label_indices = [i if i >= 0 else num_columns + i for i in label_indices]
        if all_header is not None:
            if label_names is not None:
                header = [e for e in all_header if e not in label_names]
                label_header = label_names
            else:
                header = [e for i, e in enumerate(all_header) if i not in label_indices]  # type: ignore
                label_header = [all_header[i] for i in label_indices]  # type: ignore
        else:
            if label_names:
                raise ValueError(
                    f"header is not found at `{bundle.x_train}` but `label_names` is provided. "
                    f"please either provide header in `{bundle.x_train}`, "
                    "or use `label_indices` instead."
                )
            header = []
            label_header = []
            all_header = []
            x_counter = y_counter = 0
            for i in range(num_columns):
                if i not in label_indices:  # type: ignore
                    header.append(get_x_column_name(x_counter))
                    all_header.append(header[-1])
                    x_counter += 1
                else:
                    label_header.append(get_y_column_name(y_counter))
                    all_header.append(label_header[-1])
                    y_counter += 1
        self.feature_header = header
        self.label_header = label_header
        self.all_header = all_header
        # recognize data types
        data_T = list(zip(*data))
        self.converters = {}
        for column_name, line in zip(all_header, data_T):
            is_label = column_name in label_header
            custom_mapping = self.custom_mappings.get(column_name)
            default_value = self.default_values.get(column_name)
            args = column_name, line, is_label, custom_mapping, default_value
            converter = Converter.fit(*args)  # type: ignore
            self.converters[column_name] = converter
        # transform
        train_data_T = data_T
        valid_data_T = None if valid_data is None else list(zip(*valid_data))
        return self._transform(bundle, False, (train_data_T, valid_data_T))

    def recover_labels(self, y: np.ndarray) -> np.ndarray:
        label_converters = [self.converters[h] for h in self.label_header]
        # check bypass
        if len(label_converters) != y.shape[1]:
            return y
        if len(label_converters) == 1:
            converter = label_converters[0]
            if converter.dtype == DataTypes.FLOAT:
                return y
            if is_float(y):
                return y
        recovered = []
        for i, converter in enumerate(label_converters):
            iy = y[..., i]
            if converter.dtype == DataTypes.FLOAT:
                recovered.append(iy)
            else:
                rev_mapping: Dict[int, Union[int, str]]
                if converter.dtype == DataTypes.STRING:
                    rev_mapping = {v: k for k, v in converter.mapping.items()}
                else:
                    rev_mapping = {v: int(k) for k, v in converter.mapping.items()}
                recovered.append(np.array([rev_mapping[e] for e in iy]))
        return np.ascontiguousarray(np.vstack(recovered).T)

    # internal

    def _check_file(self, x: data_type) -> bool:
        if not isinstance(x, str):
            if self.is_local_rank_0:
                print_warning("data is not a file, `FileParserBlock` will do nothing")
            return False
        return True

    def _setup_none_file(self, x: np.ndarray, y: np.ndarray) -> None:
        self.feature_header = list(map(get_x_column_name, range(x.shape[-1])))
        self.label_header = list(map(get_y_column_name, range(y.shape[-1])))
        self.all_header = self.feature_header + self.label_header
        self.converters = {}
        for h in self.feature_header:
            self.converters[h] = Converter.get_placeholder(h, False)
        for h in self.label_header:
            self.converters[h] = Converter.get_placeholder(h, True)

    def _read(self, file: str) -> Tuple[Optional[List[str]], List[List[str]]]:
        with open(file, "r") as f:
            data = list(csv.reader(f, delimiter=self.delimiter))
            for i in range(len(data) - 1, -1, -1):
                if not data[i]:
                    data.pop(i)
        if not self.has_header:
            header = None
        else:
            header = [e.strip() for e in data.pop(0)]
        return header, data

    def _convert_data(self, data_T: List[List[str]]) -> TArrayPair:
        x_list: List[np.ndarray] = []
        y_list: Optional[List[np.ndarray]] = [] if self.contain_labels else None
        header = self.all_header if self.contain_labels else self.feature_header
        for column_name, line in zip(header, data_T):
            converter = self.converters.get(column_name)
            if converter is None:
                raise ValueError(
                    f"unrecognized column name '{column_name}' occurred, "
                    f"supported column names are: {', '.join(self.all_header)}"
                )
            array = converter.convert(line, verbose=self.is_local_rank_0).ravel()
            if converter.is_label and self.auto_convert_labels:
                if converter.dtype != DataTypes.FLOAT:
                    array = array.astype(int)
            (y_list if converter.is_label else x_list).append(array)  # type: ignore
        if y_list is None:
            return np.ascontiguousarray(np.vstack(x_list).T), None
        return tuple(  # type: ignore
            map(
                np.ascontiguousarray,
                map(np.transpose, map(np.vstack, [x_list, y_list])),
            )
        )

    def _transform(
        self,
        bundle: DataBundle,
        for_inference: bool,
        train_valid_T: Optional[Tuple[Any, Any]] = None,
    ) -> DataBundle:
        if not self._check_file(bundle.x_train):
            return bundle
        if not for_inference and not self.contain_labels:
            raise ValueError("`contain_labels` should be True when not `for_inference`")
        if train_valid_T is not None:
            train_T, valid_T = train_valid_T
        else:
            if not isinstance(bundle.x_train, str):
                raise ValueError("`FileParserBlock` should be used on file inputs")
            train_T = list(zip(*self._read(bundle.x_train)[1]))
            if bundle.x_valid is None:
                valid_T = None
            else:
                if not isinstance(bundle.x_valid, str):
                    raise ValueError("`FileParserBlock` should be used on file inputs")
                valid_T = list(zip(*self._read(bundle.x_valid)[1]))
        if len(train_T) != len(self.all_header):
            if for_inference and len(train_T) + 1 == len(self.all_header):
                if self.is_local_rank_0:
                    print_info(
                        "labels are not detected and `for_inference` is set to True, "
                        "so `contain_labels` will be set to False"
                    )
                self.contain_labels = False
            else:
                raise ValueError(
                    f"number of columns ({len(train_T)}) does not match "
                    f"current number of columns ({len(self.all_header)})"
                )
        x_train, y_train = self._convert_data(train_T)
        if valid_T is None:
            x_valid = y_valid = None
        else:
            x_valid, y_valid = self._convert_data(valid_T)
        return DataBundle(x_train, y_train, x_valid, y_valid)


__all__ = [
    "MLFileProcessorConfig",
    "FileParserBlock",
]



import numpy as np

from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Optional
from typing import NamedTuple
from dataclasses import dataclass
from cftool.misc import print_warning
from cftool.array import is_float
from cftool.array import get_unique_indices

from .recognizer import RecognizerBlock
from ....schema import DataBundle
from ....schema import ColumnTypes
from ....schema import INoInitDataBlock


num_split_type = Union[int, float]
xy_type = Tuple[np.ndarray, np.ndarray]


def get_num(n: num_split_type, total: int) -> int:
    if n < 1.0 or (n == 1.0 and isinstance(n, float)):
        return int(round(total * n))
    return int(n)


def preprocess_labels(y: np.ndarray) -> np.ndarray:
    if len(y.shape) == 1:
        y = y[..., None]
    return y


class DataOrder(str, Enum):
    NONE = "none"
    TOP_DOWN = "top_down"
    BOTTOM_UP = "bottom_up"


class DataSplit(NamedTuple):
    x: np.ndarray
    y: np.ndarray
    x_remained: np.ndarray
    y_remained: np.ndarray
    split_indices: np.ndarray
    remained_indices: np.ndarray

    @property
    def xy(self) -> xy_type:
        return self.x, self.y

    @property
    def xy_remained(self) -> xy_type:
        return self.x_remained, self.y_remained


class DataSplitter:
    """
    Util class for dividing dataset based on task type
    * If it's regression task, it's not difficult to split data
    * If it's classification task, we need to split data based on the labels, because we need to:
      * ensure the divided data contain all labels available
      * ensure the label distribution is identical across all divided data.

    Examples
    --------
    >>> import numpy as np
    >>>
    >>> from cflearn.data.ml.utils import DataSplitter
    >>>
    >>> x = np.arange(12).reshape([6, 2])
    >>> # create an imbalance dataset
    >>> y = np.zeros(6, int)
    >>> y[[-1, -2]] = 1
    >>> splitter = DataSplitter()
    >>> # labels in result will keep its ratio
    >>> result = splitter.split(x, y, 3)
    >>> print(result.y.ravel())  # [0 0 1]
    >>> result = splitter.split(x, y, 0.5)
    >>> print(result.y.ravel())  # [0 0 1]
    >>> # at least one sample of each class will be kept
    >>> y[-2] = 0
    >>> result = splitter.split(x, y, 2)
    >>> # [0 0 0 0 0 1] [0 1]
    >>> print(y, result.dataset.y.ravel())

    """

    def __init__(
        self,
        *,
        order: DataOrder = DataOrder.NONE,
        shuffle: bool = True,
        verbose: bool = True,
    ) -> None:
        if shuffle and order != DataOrder.NONE:
            raise ValueError(
                "`order` should be set to 'none' "
                f"when `shuffle` is set to 'True', but '{order}' found"
            )
        self._order = order
        self._shuffle = shuffle
        self._verbose = verbose

    # internal

    def _split_with_indices(
        self,
        x: np.ndarray,
        y: np.ndarray,
        split_indices: np.ndarray,
        remained_indices: np.ndarray,
    ) -> DataSplit:
        return DataSplit(
            x[split_indices],
            y[split_indices],
            x[remained_indices],
            y[remained_indices],
            split_indices,
            remained_indices,
        )

    # api

    @property
    def order(self) -> DataOrder:
        return self._order

    @order.setter
    def order(self, value: DataOrder) -> None:
        self._order = value
        if self._shuffle and value != DataOrder.NONE:
            if self._verbose:
                print_warning(
                    f"`order` is set to '{value}', "
                    "so `shuffle` will be set to 'False' by force"
                )
            self._shuffle = False

    @property
    def shuffle(self) -> bool:
        return self._shuffle

    @shuffle.setter
    def shuffle(self, value: bool) -> None:
        self._shuffle = value
        if value and self._order != DataOrder.NONE:
            if self._verbose:
                print_warning(
                    f"`shuffle` is set to 'True', "
                    "so `order` will be set to 'none' by force"
                )
            self._order = DataOrder.NONE

    # will treat float `y` as regression datasets and int `y` as classification datasets
    def split(self, x: np.ndarray, y: np.ndarray, n: num_split_type) -> DataSplit:
        method = "reg" if is_float(y) else "clf"
        return getattr(self, f"split_{method}")(x, y, n)

    # split regression datasets
    def split_reg(self, x: np.ndarray, y: np.ndarray, n: num_split_type) -> DataSplit:
        y = preprocess_labels(y)
        total = len(x)
        n = get_num(n, total)
        if self._shuffle:
            base = np.random.permutation(total)
        else:
            base = np.arange(total)
            if self._order == DataOrder.BOTTOM_UP:
                base = np.ascontiguousarray(base[::-1])
        split_indices = base[:n]
        remained_indices = base[n:]
        return self._split_with_indices(x, y, split_indices, remained_indices)

    # split classification datasets
    def split_clf(self, x: np.ndarray, y: np.ndarray, n: num_split_type) -> DataSplit:
        y = preprocess_labels(y)
        total = len(x)
        n = get_num(n, total)
        res = get_unique_indices(y.ravel())
        label_counts = res.unique_cnt
        label_ratios = label_counts.astype(np.float64) / total
        unique_labels = res.unique.astype(int)
        num_unique_labels = len(unique_labels)
        if n < num_unique_labels:
            raise ValueError(
                f"at least {num_unique_labels} samples are required because "
                f"we have {num_unique_labels} unique labels"
            )
        label_split_indices = [indices.astype(int) for indices in res.split_indices]
        if num_unique_labels == 1:
            raise ValueError(
                "only 1 unique label is detected, "
                "which is invalid in classification task"
            )
        if self._shuffle:
            tuple(map(np.random.shuffle, label_split_indices))
        elif self._order == DataOrder.BOTTOM_UP:
            label_split_indices = [indices[::-1] for indices in label_split_indices]
            label_split_indices = list(map(np.ascontiguousarray, label_split_indices))
        rounded = np.round(n * label_ratios).astype(int)
        num_samples_per_label = np.maximum(1, rounded)
        # -num_unique_labels <= num_samples_exceeded <= num_unique_labels
        num_samples_exceeded = num_samples_per_label.sum() - n
        # adjust num_samples_per_label to make sure `n` samples are split out
        if num_samples_exceeded != 0:
            sign = np.sign(num_samples_exceeded)
            num_samples_exceeded = abs(num_samples_exceeded)
            arange = np.arange(num_unique_labels)
            chosen_indices = arange[num_samples_per_label != 1]
            np.random.shuffle(chosen_indices)
            num_chosen_indices = len(chosen_indices)
            num_tile = int(np.ceil(num_samples_exceeded / num_chosen_indices))
            num_proceeded = 0
            for _ in range(num_tile - 1):
                num_samples_per_label[chosen_indices] -= sign
                num_proceeded += num_chosen_indices
            for idx in chosen_indices[: num_samples_exceeded - num_proceeded]:
                num_samples_per_label[idx] -= sign
        assert num_samples_per_label.sum() == n
        num_overlap = 0
        split_indices_list: List[np.ndarray] = []
        remained_indices_list: List[np.ndarray] = []
        for indices, num_sample_per_label in zip(
            label_split_indices,
            num_samples_per_label,
        ):
            num_samples_available = len(indices)
            split_indices_list.append(indices[:num_sample_per_label])
            if num_sample_per_label >= num_samples_available:
                num_overlap += num_sample_per_label
                remained_indices_list.append(indices)
            else:
                remained_indices_list.append(indices[num_sample_per_label:])
        split_indices = np.hstack(split_indices_list)
        remained_indices = np.hstack(remained_indices_list)
        base = np.zeros(total)
        base[split_indices] += 1
        base[remained_indices] += 1
        assert np.sum(base >= 2) <= num_overlap
        return self._split_with_indices(x, y, split_indices, remained_indices)


@dataclass
class MLSplitterConfig:
    num_split: Optional[num_split_type] = None
    min_split: Optional[int] = None
    max_split: int = 10000
    split_order: DataOrder = DataOrder.NONE
    split_shuffle: bool = True
    is_classification: Optional[bool] = None


@INoInitDataBlock.register("ml_splitter")
class SplitterBlock(INoInitDataBlock):
    config: MLSplitterConfig  # type: ignore
    num_split: Optional[num_split_type]
    min_split: Optional[int]
    max_split: int
    order: DataOrder
    shuffle: bool
    is_classification: bool

    # inheritance

    def to_info(self) -> Dict[str, Any]:
        return dict(
            num_split=self.num_split,
            min_split=self.min_split,
            max_split=self.max_split,
            order=self.order,
            shuffle=self.shuffle,
            is_classification=self.is_classification,
        )

    def transform(self, bundle: DataBundle, for_inference: bool) -> DataBundle:
        if for_inference or bundle.y_train is None or bundle.x_valid is not None:
            return bundle
        if not isinstance(bundle.x_train, np.ndarray):
            return bundle

        if self.num_split is None:
            num = 0.1
            default_min_split = 100
        else:
            num = self.num_split
            default_min_split = 0
        total = len(bundle.x_train)
        if isinstance(num, float):
            num = round(total * num)
        if num <= 0:
            return bundle
        min_split = default_min_split if self.min_split is None else self.min_split
        num = min(self.max_split, max(min_split, num))
        if num == min_split and num >= 0.5 * total:
            print_warning(
                f"`min_split` ({min_split}) exceeds "
                f"half of the total number of samples ({total}), "
                "so `SplitterBlock` will take no effects to "
                "ensure we have enought training samples."
            )
            return bundle
        kw = dict(order=self.order, shuffle=self.shuffle, verbose=self.is_local_rank_0)
        splitter = DataSplitter(**kw)  # type: ignore
        fn = splitter.split_clf if self.is_classification else splitter.split_reg
        split = fn(bundle.x_train, bundle.y_train, num)
        return DataBundle(*split.xy_remained, *split.xy)

    def fit_transform(self, bundle: DataBundle) -> DataBundle:
        self.num_split = self.config.num_split
        self.min_split = self.config.min_split
        self.max_split = self.config.max_split
        self.order = self.config.split_order
        self.shuffle = self.config.split_shuffle
        is_classification = self.config.is_classification
        if is_classification is None:
            if self.recognizer is None:
                is_classification = not is_float(bundle.y_train)
            else:
                is_classification = all(
                    t == ColumnTypes.CATEGORICAL
                    for t in self.recognizer.label_types.values()
                )
        self.is_classification = self.config.is_classification = is_classification
        return self.transform(bundle, False)

    # api

    @property
    def recognizer(self) -> Optional[RecognizerBlock]:
        return self.try_get_previous(RecognizerBlock)


__all__ = [
    "DataOrder",
    "DataSplitter",
    "MLSplitterConfig",
    "SplitterBlock",
]



from torch import Tensor
from typing import Any
from typing import Dict
from typing import Type
from typing import Optional
from cftool.misc import safe_execute
from cftool.misc import print_warning
from cftool.misc import shallow_copy_dict
from cftool.misc import WithRegister
from cftool.types import tensor_dict_type

from ....schema import _forward
from ....schema import TrainerState
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY
from ....constants import PREDICTIONS_KEY
from ....misc.toolkit import interpolate
from ....misc.toolkit import auto_num_layers
from ....modules.blocks import ChannelPadding


decoders: Dict[str, Type["DecoderMixin"]] = {}
decoders_1d: Dict[str, Type["Decoder1DMixin"]] = {}


class IDecoder:
    num_upsample: int
    out_channels: int
    img_size: Optional[int]
    num_classes: Optional[int]
    latent_resolution: Optional[int]

    def _initialize(
        self,
        *,
        out_channels: int,
        img_size: Optional[int],
        num_upsample: Optional[int],
        num_classes: Optional[int],
        latent_resolution: Optional[int],
    ) -> None:
        if num_upsample is None:
            fmt = "`{}` should be provided when `num_upsample` is not"
            if img_size is None:
                raise ValueError(fmt.format("img_size"))
            if latent_resolution is None:
                print_warning(
                    f'{fmt.format("latent_resolution")}, '
                    "and 7 will be used as the default `latent_resolution` now"
                )
                latent_resolution = 7
            num_upsample = auto_num_layers(img_size, latent_resolution, None)
        self.num_upsample = num_upsample
        self.out_channels = out_channels
        self.img_size = img_size
        self.num_classes = num_classes
        self.latent_resolution = latent_resolution

    @property
    def is_conditional(self) -> bool:
        return self.num_classes is not None

    def resize(self, net: Tensor, *, determinate: bool = False) -> Tensor:
        if self.img_size is None:
            return net
        return interpolate(net, size=self.img_size, determinate=determinate)

    def decode(self, batch: tensor_dict_type, **kwargs: Any) -> Tensor:
        return run_decoder(self, 0, batch, **kwargs)[PREDICTIONS_KEY]


# decode from latent feature map
class DecoderMixin(IDecoder, WithRegister["DecoderMixin"]):
    d = decoders

    latent_channels: int
    cond_channels: int

    def _init_cond(self, *, cond_channels: int = 16) -> None:
        self.cond_channels = cond_channels
        if self.num_classes is None:
            self.cond = None
        else:
            if self.latent_resolution is None:
                msg = "`latent_resolution` should be provided for conditional modeling"
                raise ValueError(msg)
            self.cond = ChannelPadding(
                self.latent_channels,
                cond_channels,
                self.latent_resolution,
                num_classes=self.num_classes,
            )

    def _inject_cond(self, batch: tensor_dict_type) -> tensor_dict_type:
        if self.cond is None:
            return batch
        batch = shallow_copy_dict(batch)
        batch[INPUT_KEY] = self.cond(batch[INPUT_KEY], batch.get(LABEL_KEY))
        return batch


# decode from 1d latent code
class Decoder1DMixin(IDecoder, WithRegister["Decoder1DMixin"]):
    d = decoders_1d

    latent_dim: int


def make_decoder(name: str, config: Dict[str, Any], *, is_1d: bool = False) -> IDecoder:
    base = (Decoder1DMixin if is_1d else DecoderMixin).get(name)  # type: ignore
    return safe_execute(base, config)


def run_decoder(
    decoder: IDecoder,
    batch_idx: int,
    batch: tensor_dict_type,
    state: Optional[TrainerState] = None,
    **kwargs: Any,
) -> tensor_dict_type:
    return _forward(decoder, batch_idx, batch, INPUT_KEY, state, **kwargs)


__all__ = [
    "make_decoder",
    "run_decoder",
    "IDecoder",
    "DecoderMixin",
    "Decoder1DMixin",
]



import os
import torch
import shutil

import numpy as np

from abc import abstractmethod
from abc import ABCMeta
from enum import Enum
from tqdm import tqdm
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Union
from typing import TypeVar
from typing import Callable
from typing import Optional
from tempfile import mkdtemp
from tempfile import TemporaryDirectory
from collections import OrderedDict
from cftool.misc import print_info
from cftool.misc import safe_execute
from cftool.misc import shallow_copy_dict
from cftool.misc import prepare_workspace_from
from cftool.misc import Saving
from cftool.misc import Serializer
from cftool.array import sigmoid
from cftool.array import softmax
from cftool.types import np_dict_type
from cftool.types import tensor_dict_type
from cftool.pipeline import get_workspace

from .core import Block
from .core import Pipeline
from .blocks import SetDefaultsBlock
from .blocks import SetMLDefaultsBlock
from .blocks import PrepareWorkplaceBlock
from .blocks import ExtractStateInfoBlock
from .blocks import BuildLossBlock
from .blocks import BuildModelBlock
from .blocks import BuildMetricsBlock
from .blocks import BuildInferenceBlock
from .blocks import SetTrainerDefaultsBlock
from .blocks import SetMLTrainerDefaultsBlock
from .blocks import BuildMonitorsBlock
from .blocks import BuildCallbacksBlock
from .blocks import BuildOptimizersBlock
from .blocks import BuildTrainerBlock
from .blocks import RecordNumSamplesBlock
from .blocks import ReportBlock
from .blocks import TrainingBlock
from .blocks import SerializeDataBlock
from .blocks import SerializeModelBlock
from .blocks import SerializeOptimizerBlock
from .schema import IEvaluationPipeline
from ..types import sample_weights_type
from ..types import states_callback_type
from ..schema import IData
from ..schema import DLConfig
from ..schema import IDataLoader
from ..schema import MetricsOutputs
from ..trainer import get_scores
from ..trainer import get_metrics
from ..trainer import get_input_sample
from ..trainer import get_sorted_checkpoints
from ..constants import PREDICTIONS_KEY
from ..data.ml import MLData
from ..misc.toolkit import is_local_rank_0
from ..misc.toolkit import get_device


TInferPipeline = TypeVar("TInferPipeline", bound="DLInferencePipeline", covariant=True)


# internal mixins


class _DeviceMixin:
    build_model: BuildModelBlock

    @property
    def device(self) -> torch.device:
        return self.build_model.model.device


class _InferenceMixin:
    focuses: List[Type[Block]]
    is_built: bool

    data: Optional[IData]
    get_block: Callable[[Type[Block]], Any]
    try_get_block: Callable[[Type[Block]], Any]

    # optional callbacks

    def predict_callback(self, results: np_dict_type) -> np_dict_type:
        """changes can happen inplace"""
        return results

    # api

    @property
    def build_model(self) -> BuildModelBlock:
        return self.get_block(BuildModelBlock)

    @property
    def build_inference(self) -> BuildInferenceBlock:
        return self.get_block(BuildInferenceBlock)

    @property
    def serialize_data(self) -> SerializeDataBlock:
        return self.get_block(SerializeDataBlock)

    @property
    def serialize_model(self) -> Optional[SerializeModelBlock]:
        return self.try_get_block(SerializeModelBlock)

    @classmethod
    def build_with(  # type: ignore
        cls: Type[TInferPipeline],
        config: DLConfig,
        states: Optional[tensor_dict_type] = None,
        *,
        data: Optional[IData] = None,
    ) -> TInferPipeline:
        self = cls.init(config)
        # last focus will be the serialization block
        self.build(*[Block.make(b.__identifier__, {}) for b in cls.focuses])
        if states is not None:
            self.build_model.model.load_state_dict(states)
        self.serialize_model.verbose = False
        self.serialize_data.data = self.data = data
        self.is_built = True
        return self

    def to(  # type: ignore
        self: TInferPipeline,
        device: Union[int, str, torch.device],
    ) -> TInferPipeline:
        self.build_model.model.to(device)
        return self

    def predict(
        self,
        loader: IDataLoader,
        *,
        return_classes: bool = False,
        binary_threshold: float = 0.5,
        return_probabilities: bool = False,
        recover_labels: Optional[bool] = None,
        **kwargs: Any,
    ) -> np_dict_type:
        if not self.is_built:
            raise ValueError(
                f"`{self.__class__.__name__}` should be built beforehand, please use "
                "`DLPipelineSerializer.load_inference/evaluation` or `build_with` "
                "to get a built one!"
            )
        kw = shallow_copy_dict(kwargs)
        kw["loader"] = loader
        outputs = safe_execute(self.build_inference.inference.get_outputs, kw)
        results = outputs.forward_results
        # handle predict flags
        if return_classes and return_probabilities:
            raise ValueError(
                "`return_classes` & `return_probabilities`"
                "should not be True at the same time"
            )
        elif not return_classes and not return_probabilities:
            pass
        else:
            predictions = results[PREDICTIONS_KEY]
            if predictions.shape[1] > 2 and return_classes:
                results[PREDICTIONS_KEY] = predictions.argmax(1, keepdims=True)
            else:
                if predictions.shape[1] == 2:
                    probabilities = softmax(predictions)
                else:
                    pos = sigmoid(predictions)
                    probabilities = np.hstack([1.0 - pos, pos])
                if return_probabilities:
                    results[PREDICTIONS_KEY] = probabilities
                else:
                    classes = (probabilities[..., [1]] >= binary_threshold).astype(int)
                    results[PREDICTIONS_KEY] = classes
        # handle recover labels
        if recover_labels is None:
            recover_labels = self.data is not None
        if recover_labels:
            if self.data is None:
                msg = "`recover_labels` is set to `True` but `data` is not provided"
                raise ValueError(msg)
            y = results[PREDICTIONS_KEY]
            results[PREDICTIONS_KEY] = self.data.recover_labels(y)
        # optional callback
        results = self.predict_callback(results)
        # return
        return results


class _EvaluationMixin(_InferenceMixin, IEvaluationPipeline):
    config: DLConfig

    @property
    def build_loss(self) -> BuildLossBlock:
        return self.get_block(BuildLossBlock)

    @property
    def build_metrics(self) -> BuildMetricsBlock:
        return self.get_block(BuildMetricsBlock)

    def evaluate(self, loader: IDataLoader, **kwargs: Any) -> MetricsOutputs:
        return get_metrics(
            self.config,
            self.build_model.model,
            self.build_loss.loss,
            self.build_metrics.metrics,
            self.build_inference.inference,
            loader,
            forward_kwargs=kwargs,
        )


# apis


class PipelineTypes(str, Enum):
    DL_TRAINING = "dl.training"
    ML_TRAINING = "ml.training"
    DL_INFERENCE = "dl.inference"
    DL_EVALUATION = "dl.evaluation"


class TrainingPipeline(
    Pipeline,
    _DeviceMixin,
    _EvaluationMixin,
    metaclass=ABCMeta,
):
    is_built = False

    @property
    @abstractmethod
    def set_defaults_block(self) -> Block:
        pass

    @property
    @abstractmethod
    def set_trainer_defaults_block(self) -> Block:
        pass

    @property
    def build_trainer(self) -> BuildTrainerBlock:
        return self.get_block(BuildTrainerBlock)

    @property
    def building_blocks(self) -> List[Block]:
        return [
            self.set_defaults_block,
            PrepareWorkplaceBlock(),
            ExtractStateInfoBlock(),
            BuildLossBlock(),
            BuildModelBlock(),
            BuildMetricsBlock(),
            BuildInferenceBlock(),
            self.set_trainer_defaults_block,
            BuildMonitorsBlock(),
            BuildCallbacksBlock(),
            BuildOptimizersBlock(),
            BuildTrainerBlock(),
            RecordNumSamplesBlock(),
            ReportBlock(),
            TrainingBlock(),
            SerializeDataBlock(),
            SerializeModelBlock(),
            SerializeOptimizerBlock(),
        ]

    def after_load(self) -> None:
        self.is_built = True
        workspace = prepare_workspace_from("_logs")
        self.config.workspace = workspace

    def prepare(self, data: IData, sample_weights: sample_weights_type = None) -> None:
        self.data = data.set_sample_weights(sample_weights)
        self.training_workspace = self.config.workspace
        if not self.is_built:
            self.build(*self.building_blocks)
            self.is_built = True
        else:
            for block in self.blocks:
                block.training_workspace = self.training_workspace

    def fit(
        self,
        data: IData,
        *,
        sample_weights: sample_weights_type = None,
        cuda: Optional[Union[int, str]] = None,
    ) -> "TrainingPipeline":
        # build pipeline
        self.prepare(data, sample_weights)
        # check rank 0
        workspace = self.config.workspace if is_local_rank_0() else None
        # save data info
        if workspace is not None:
            Serializer.save(
                os.path.join(workspace, SerializeDataBlock.package_folder),
                data,
                save_npd=False,
            )
        # run pipeline
        self.run(data, cuda=cuda)
        # save pipeline
        if workspace is not None:
            pipeline_folder = DLPipelineSerializer.pipeline_folder
            DLPipelineSerializer.save(self, os.path.join(workspace, pipeline_folder))
        # return
        return self


@Pipeline.register(PipelineTypes.DL_TRAINING)
class DLTrainingPipeline(TrainingPipeline):
    @property
    def set_defaults_block(self) -> Block:
        return SetDefaultsBlock()

    @property
    def set_trainer_defaults_block(self) -> Block:
        return SetTrainerDefaultsBlock()


@Pipeline.register(PipelineTypes.ML_TRAINING)
class MLTrainingPipeline(TrainingPipeline):
    data: MLData

    @property
    def set_defaults_block(self) -> Block:
        return SetMLDefaultsBlock()

    @property
    def set_trainer_defaults_block(self) -> Block:
        return SetMLTrainerDefaultsBlock()


@Pipeline.register(PipelineTypes.DL_INFERENCE)
class DLInferencePipeline(Pipeline, _DeviceMixin, _InferenceMixin):
    is_built = False

    focuses = [
        BuildModelBlock,
        BuildInferenceBlock,
        SerializeDataBlock,
        SerializeModelBlock,
    ]

    def after_load(self) -> None:
        self.is_built = True
        self.data = self.serialize_data.data
        if self.serialize_model is not None:
            self.serialize_model.verbose = False


@Pipeline.register(PipelineTypes.DL_EVALUATION)
class DLEvaluationPipeline(DLInferencePipeline, _EvaluationMixin):
    focuses = [
        BuildLossBlock,
        BuildModelBlock,
        BuildMetricsBlock,
        BuildInferenceBlock,
        SerializeDataBlock,
        SerializeModelBlock,
    ]


class PackType(str, Enum):
    TRAINING = "training"
    INFERENCE = "inference"
    EVALUATION = "evaluation"


class DLPipelineSerializer:
    id_file = "id.txt"
    config_file = "config.json"
    blocks_file = "blocks.json"
    pipeline_folder = "pipeline"

    # api

    @classmethod
    def save(cls, pipeline: Pipeline, folder: str, *, compress: bool = False) -> None:
        original_folder = None
        if compress:
            original_folder = folder
            folder = mkdtemp()
        Serializer.save(folder, pipeline)
        for block in pipeline.blocks:
            block.save_extra(os.path.join(folder, block.__identifier__))
        if compress and original_folder is not None:
            abs_folder = os.path.abspath(folder)
            abs_original = os.path.abspath(original_folder)
            Saving.compress(abs_folder)
            shutil.move(f"{abs_folder}.zip", f"{abs_original}.zip")

    @classmethod
    def pack(
        cls,
        workspace: str,
        export_folder: str,
        *,
        pack_type: PackType = PackType.INFERENCE,
        compress: bool = True,
    ) -> None:
        if pack_type == PackType.TRAINING:
            swap_id = None
            focuses = None
            excludes = [PrepareWorkplaceBlock]
        elif pack_type == PackType.INFERENCE:
            swap_id = DLInferencePipeline.__identifier__
            focuses = DLInferencePipeline.focuses
            excludes = None
        elif pack_type == PackType.EVALUATION:
            swap_id = DLEvaluationPipeline.__identifier__
            focuses = DLEvaluationPipeline.focuses
            excludes = None
        else:
            raise ValueError(f"unrecognized `pack_type` '{pack_type}' occurred")
        pipeline_folder = os.path.join(workspace, cls.pipeline_folder)
        pipeline = cls._load(
            pipeline_folder,
            swap_id=swap_id,
            focuses=focuses,
            excludes=excludes,
        )
        cls.save(pipeline, export_folder, compress=compress)

    @classmethod
    def pack_and_load_inference(cls, workplace: str) -> DLInferencePipeline:
        with TemporaryDirectory() as tmp_folder:
            cls.pack(
                workplace,
                export_folder=tmp_folder,
                pack_type=PackType.INFERENCE,
                compress=False,
            )
            return cls.load_inference(tmp_folder)

    @classmethod
    def pack_onnx(
        cls,
        workplace: str,
        export_file: str = "model.onnx",
        dynamic_axes: Optional[Union[List[int], Dict[int, str]]] = None,
        *,
        input_sample: Optional[tensor_dict_type] = None,
        loader_sample: Optional[IDataLoader] = None,
        opset: int = 11,
        simplify: bool = True,
        num_samples: Optional[int] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> DLInferencePipeline:
        if input_sample is None and loader_sample is None:
            msg = "either `input_sample` or `loader_sample` should be provided"
            raise ValueError(msg)
        m = cls.pack_and_load_inference(workplace)
        model = m.build_model.model
        if input_sample is None:
            input_sample = get_input_sample(loader_sample, get_device(model))  # type: ignore
        model.to_onnx(
            export_file,
            input_sample,
            dynamic_axes,
            opset=opset,
            simplify=simplify,
            num_samples=num_samples,
            verbose=verbose,
            **kwargs,
        )
        return m

    @classmethod
    def pack_scripted(
        cls,
        workplace: str,
        export_file: str = "model.pt",
    ) -> DLInferencePipeline:
        m = cls.pack_and_load_inference(workplace)
        model = torch.jit.script(m.build_model.model)
        torch.jit.save(model, export_file)
        return m

    @classmethod
    def fuse_inference(
        cls,
        src_folders: List[str],
        *,
        cuda: Optional[str] = None,
        num_picked: Optional[Union[int, float]] = None,
        states_callback: states_callback_type = None,
    ) -> DLInferencePipeline:
        return cls._fuse_multiple(
            src_folders,
            PackType.INFERENCE,
            cuda,
            num_picked,
            states_callback,
        )

    @classmethod
    def fuse_evaluation(
        cls,
        src_folders: List[str],
        *,
        cuda: Optional[str] = None,
        num_picked: Optional[Union[int, float]] = None,
        states_callback: states_callback_type = None,
    ) -> DLEvaluationPipeline:
        return cls._fuse_multiple(
            src_folders,
            PackType.EVALUATION,
            cuda,
            num_picked,
            states_callback,
        )

    @classmethod
    def load_training(cls, folder: str) -> TrainingPipeline:
        return cls._load(folder, swap_id=DLTrainingPipeline.__identifier__)

    @classmethod
    def load_inference(cls, folder: str) -> DLInferencePipeline:
        return cls._load_inference(folder)

    @classmethod
    def load_evaluation(cls, folder: str) -> DLEvaluationPipeline:
        return cls._load_evaluation(folder)

    # internal

    @classmethod
    def _load(
        cls,
        folder: str,
        *,
        swap_id: Optional[str] = None,
        focuses: Optional[List[Type[Block]]] = None,
        excludes: Optional[List[Type[Block]]] = None,
    ) -> Pipeline:
        with get_workspace(folder) as workspace:
            # handle info
            info = Serializer.load_info(workspace)
            if focuses is not None or excludes is not None:
                if focuses is None:
                    focuses_set = None
                else:
                    focuses_set = {b.__identifier__ for b in focuses}
                block_types = info["blocks"]
                if focuses_set is not None:
                    block_types = [b for b in block_types if b in focuses_set]
                    left = sorted(focuses_set - set(block_types))
                    if left:
                        raise ValueError(
                            "following blocks are specified in `focuses` "
                            f"but not found in the loaded blocks: {', '.join(left)}"
                        )
                if excludes is not None:
                    excludes_set = {b.__identifier__ for b in excludes}
                    block_types = [b for b in block_types if b not in excludes_set]
                info["blocks"] = block_types
            # load
            pipeline = Serializer.load_empty(workspace, Pipeline, swap_id=swap_id)
            pipeline.serialize_folder = workspace
            if info is None:
                info = Serializer.load_info(workspace)
            pipeline.from_info(info)
            for block in pipeline.blocks:
                block.load_from(os.path.join(workspace, block.__identifier__))
            pipeline.after_load()
        return pipeline

    @classmethod
    def _load_inference(
        cls,
        folder: str,
        excludes: Optional[List[Type[Block]]] = None,
    ) -> DLInferencePipeline:
        return cls._load(
            folder,
            swap_id=DLInferencePipeline.__identifier__,
            focuses=DLInferencePipeline.focuses,
            excludes=excludes,
        )

    @classmethod
    def _load_evaluation(
        cls,
        folder: str,
        excludes: Optional[List[Type[Block]]] = None,
    ) -> DLEvaluationPipeline:
        return cls._load(
            folder,
            swap_id=DLEvaluationPipeline.__identifier__,
            focuses=DLEvaluationPipeline.focuses,
            excludes=excludes,
        )

    @classmethod
    def _fuse_multiple(
        cls,
        src_folders: List[str],
        pack_type: PackType,
        cuda: Optional[str] = None,
        num_picked: Optional[Union[int, float]] = None,
        states_callback: states_callback_type = None,
    ) -> DLInferencePipeline:
        if pack_type == PackType.TRAINING:
            raise ValueError("should not pack to training pipeline when fusing")
        # get num picked
        num_total = num_repeat = len(src_folders)
        if num_picked is not None:
            if isinstance(num_picked, float):
                if num_picked < 0.0 or num_picked > 1.0:
                    raise ValueError("`num_picked` should ∈ [0, 1] when set to float")
                num_picked = round(num_total * num_picked)
            if num_picked < 1:
                raise ValueError("calculated `num_picked` should be at least 1")
            scores = []
            for i, folder in enumerate(src_folders):
                ckpt_folder = os.path.join(folder, SerializeModelBlock.__identifier__)
                folder_scores = get_scores(ckpt_folder)
                scores.append(max(folder_scores.values()))
            scores_array = np.array(scores)
            picked_indices = np.argsort(scores)[::-1][:num_picked]
            src_folders = [src_folders[i] for i in picked_indices]
            original_score = scores_array.mean().item()
            picked_score = scores_array[picked_indices].mean().item()
            print_info(
                f"picked {num_picked} / {num_total}, "
                f"score: {original_score} -> {picked_score}"
            )
            num_repeat = num_picked
        # get empty pipeline
        with get_workspace(src_folders[0], force_new=True) as workspace:
            info = Serializer.load_info(workspace)
            config: DLConfig = DLConfig.from_pack(info["config"])
            config.num_repeat = num_repeat
            info["config"] = config.to_pack().asdict()
            Serializer.save_info(workspace, info=info)
            fn = (
                cls._load_inference
                if pack_type == PackType.INFERENCE
                else cls._load_evaluation
            )
            # avoid loading model because the ensembled model has different states
            m = fn(workspace, excludes=[SerializeModelBlock])
            # but we need to build the SerializeModelBlock again for further save/load
            b_serialize_model = SerializeModelBlock()
            b_serialize_model.verbose = False
            m.build(b_serialize_model)
        # merge state dict
        merged_states: OrderedDict[str, torch.Tensor] = OrderedDict()
        for i, folder in enumerate(tqdm(src_folders, desc="fuse")):
            with get_workspace(folder) as i_folder:
                ckpt_folder = os.path.join(i_folder, SerializeModelBlock.__identifier__)
                checkpoints = get_sorted_checkpoints(ckpt_folder)
                checkpoint_path = os.path.join(ckpt_folder, checkpoints[0])
                states = torch.load(checkpoint_path, map_location=cuda)
            current_keys = list(states.keys())
            for k, v in list(states.items()):
                states[f"ms.{i}.{k}"] = v
            for k in current_keys:
                states.pop(k)
            if states_callback is not None:
                states = states_callback(m, states)
            merged_states.update(states)
        # load state dict
        model = m.build_model.model
        model.to(cuda)
        model.load_state_dict(merged_states)
        return m


__all__ = [
    "IEvaluationPipeline",
    "PipelineTypes",
    "TrainingPipeline",
    "DLTrainingPipeline",
    "MLTrainingPipeline",
    "DLInferencePipeline",
    "DLEvaluationPipeline",
    "PackType",
    "DLPipelineSerializer",
]



import math
import torch

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from cftool.types import tensor_dict_type

from .schema import Encoder1DFromPatches
from .schema import Encoder2DFromPatches
from ....constants import LATENT_KEY
from ....modules.blocks import MixedStackedEncoder


@Encoder1DFromPatches.register("vit")
class ViTEncoder(Encoder1DFromPatches):
    def __init__(
        self,
        *,
        img_size: int,
        patch_size: int,
        in_channels: int,
        latent_dim: int = 384,
        to_patches_type: str = "vanilla",
        to_patches_config: Optional[Dict[str, Any]] = None,
        num_layers: int = 12,
        dropout: float = 0.0,
        drop_path_rate: float = 0.0,
        norm_type: Optional[str] = "layer",
        norm_kwargs: Optional[Dict[str, Any]] = None,
        embedding_norm: Optional[torch.nn.Module] = None,
        residual_after_norm: bool = False,
        feedforward_dim_ratio: float = 4.0,
        attention_kwargs: Optional[Dict[str, Any]] = None,
        feedforward_kwargs: Optional[Dict[str, Any]] = None,
        use_head_token: bool = True,
        head_pooler: Optional[str] = "mean",
        use_positional_encoding: bool = True,
        norm_after_head: bool = False,
        output_dim: Optional[int] = None,
        aux_heads: Optional[List[str]] = None,
    ):
        super().__init__(
            img_size=img_size,
            patch_size=patch_size,
            in_channels=in_channels,
            latent_dim=latent_dim,
            to_patches_type=to_patches_type,
            to_patches_config=to_patches_config,
        )
        if attention_kwargs is None:
            attention_kwargs = {}
        attention_kwargs.setdefault("bias", True)
        attention_kwargs.setdefault("num_heads", latent_dim // 64)
        self.aux_heads = aux_heads
        self.encoder = MixedStackedEncoder(
            latent_dim,
            self.num_patches,
            token_mixing_type="attention",
            token_mixing_config=attention_kwargs,
            channel_mixing_config=feedforward_kwargs,
            num_layers=num_layers,
            dropout=dropout,
            drop_path_rate=drop_path_rate,
            norm_type=norm_type,
            norm_kwargs=norm_kwargs,
            embedding_norm=embedding_norm,
            residual_after_norm=residual_after_norm,
            latent_dim_ratio=feedforward_dim_ratio,
            head_pooler=head_pooler,
            use_head_token=use_head_token,
            use_positional_encoding=use_positional_encoding,
            is_vision_positional_encoding=True,
            norm_after_head=norm_after_head,
            aux_heads=aux_heads,
        )
        if output_dim is None:
            self.output_projection = None
        else:
            init = (latent_dim**-0.5) * torch.randn(latent_dim, output_dim)
            self.output_projection = torch.nn.Parameter(init)

    def forward(self, net: Tensor, **kwargs: Any) -> tensor_dict_type:
        latent = super().forward(net, **kwargs)
        rs = {LATENT_KEY: latent}
        if self.aux_heads is None:
            if self.output_projection is not None:
                rs[LATENT_KEY] = latent @ self.output_projection
        else:
            for i, k in enumerate(self.aux_heads + [LATENT_KEY]):
                rs[k] = latent[:, i]
        return rs


@Encoder2DFromPatches.register("vit")
class ViTEncoder2D(Encoder2DFromPatches):
    def __init__(
        self,
        *,
        img_size: int,
        patch_size: int,
        in_channels: int,
        latent_channels: int = 384,
        to_patches_type: str = "vanilla",
        to_patches_config: Optional[Dict[str, Any]] = None,
        num_layers: int = 12,
        dropout: float = 0.0,
        drop_path_rate: float = 0.0,
        norm_type: Optional[str] = "layer",
        norm_kwargs: Optional[Dict[str, Any]] = None,
        embedding_norm: Optional[torch.nn.Module] = None,
        residual_after_norm: bool = False,
        feedforward_dim_ratio: float = 4.0,
        attention_kwargs: Optional[Dict[str, Any]] = None,
        feedforward_kwargs: Optional[Dict[str, Any]] = None,
        use_positional_encoding: bool = True,
        norm_after_head: bool = False,
    ):
        super().__init__(
            img_size=img_size,
            patch_size=patch_size,
            in_channels=in_channels,
            latent_channels=latent_channels,
            to_patches_type=to_patches_type,
            to_patches_config=to_patches_config,
        )
        if attention_kwargs is None:
            attention_kwargs = {}
        attention_kwargs.setdefault("bias", True)
        attention_kwargs.setdefault("num_heads", latent_channels // 64)
        self.encoder = MixedStackedEncoder(
            latent_channels,
            self.num_patches,
            token_mixing_type="attention",
            token_mixing_config=attention_kwargs,
            channel_mixing_config=feedforward_kwargs,
            num_layers=num_layers,
            dropout=dropout,
            drop_path_rate=drop_path_rate,
            norm_type=norm_type,
            norm_kwargs=norm_kwargs,
            embedding_norm=embedding_norm,
            residual_after_norm=residual_after_norm,
            latent_dim_ratio=feedforward_dim_ratio,
            head_pooler=None,
            use_head_token=False,
            use_positional_encoding=use_positional_encoding,
            is_vision_positional_encoding=True,
            norm_after_head=norm_after_head,
        )

    def forward(self, net: Tensor, **kwargs: Any) -> Tensor:
        latent = super().forward(net, **kwargs)
        resolution = int(round(math.sqrt(self.num_patches)))
        latent = latent.view(latent.shape[0], resolution, resolution, -1)
        latent = latent.permute(0, 3, 1, 2).contiguous()
        return latent


__all__ = [
    "ViTEncoder",
    "ViTEncoder2D",
]



import torch.nn as nn

from abc import abstractmethod
from abc import ABC
from abc import ABCMeta
from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from torch.optim import Optimizer
from cftool.types import tensor_dict_type
from torch.cuda.amp import autocast

from ...schema import IDLModel
from ...schema import ITrainer
from ...schema import IDataLoader
from ...schema import StepOutputs
from ...schema import TrainerState
from ...schema import PrecisionType
from ...schema import TrainerConfig
from ...schema import MetricsOutputs
from ...constants import LOSS_KEY
from ...data.utils import TensorBatcher
from ...misc.toolkit import get_device
from ...misc.toolkit import no_grad_context
from ...misc.toolkit import toggle_optimizer


def get_update_fn(trainer: ITrainer) -> Callable[[Tensor, Optimizer, bool], None]:
    def update_fn(loss: Tensor, optimizer: Optimizer, update: bool) -> None:
        accelerator = trainer.accelerator
        accelerator.backward(loss)
        if update:
            trainer.clip_norm_step()
            optimizer.step()
            optimizer.zero_grad()

    return update_fn


def run_train_steps(
    m: "ModelWithCustomSteps",
    train_steps: List["CustomTrainStep"],
    *,
    batch_idx: int,
    batch: tensor_dict_type,
    trainer: "ITrainer",
    forward_kwargs: Dict[str, Any],
    loss_kwargs: Dict[str, Any],
) -> StepOutputs:
    """
    Runs a series of custom training steps on a given model and batch of data.

    Parameters
    ----------
    m : ModelWithCustomSteps
        The model to train.
    train_steps : List[CustomTrainStep]
        The custom training steps to run.
    batch_idx : int
        The current batch index.
    batch : tensor_dict_type
        The batch of data to use for training.
    trainer : ITrainer
        The trainer object used to train the model.
    forward_kwargs : Dict[str, Any]
        Additional arguments to pass to the forward pass of the model.
    loss_kwargs : Dict[str, Any]
        Additional arguments to pass to the loss function of each training step.

    Returns
    -------
    StepOutputs
        An object containing the outputs of the forward pass and the calculated loss values of the training steps.

    Step by step explanation
    ------------------------
    1. Initialize variables: `forward` (an empty dictionary), `loss_dict` (an empty dictionary), `any_update`
    (a bool flag set to `False`), `performed_scheduler_step` (a bool flag set to `False`), and `update_fn` (a
    function returned by the `get_update_fn` function defined above).
    2. Check whether the forward pass should have gradients (`fw_has_grad`) and which training step to use for the
    forward pass (`fw_train_step`). This is done by looping through each training step and checking its
    `requires_new_forward` and `requires_grad_in_forward` attributes.
    3. If `fw_has_grad` is `False` and a subsequent training step requires gradients in the forward pass, raise a
    ValueError with a message indicating which training steps have conflicting requirements.
    4. Loop through each training step and execute the following steps for each:
      1) Check whether the current training step should be skipped. If so, move on to the next training step.
      2) If this is the first training step, or if `requires_new_forward` is `True` for the current training step,
      execute the forward pass of the model and store the output in `forward`. The `no_grad_context` context manager
      is used to prevent gradients from being calculated if `requires_grad_in_forward` is `False`.
      3) Get the optimizer to be used for this training step.
      4) If `enable_toggle_optimizer` is `True` for this training step, temporarily switch to the optimizer associated
      with this training step using the `toggle_optimizer` context manager.
      5) Calculate the loss for this training step using the model, state, batch, and forward pass outputs. The
      `autocast` context manager is used if mixed-precision training is enabled.
      6) Update the optimizer if `train_step.grad_accumulate` is a factor of the current `state.step`.
      7) Update the `loss_dict` with the loss values for this training step.
      8) If an optimizer update occurred, set `any_update` to `True`, and if `requires_scheduler_step` is `True` for
      this training step, call `trainer.scheduler_step()` to update the learning rate.
    5. If any optimizer updates occurred but no scheduler steps were performed, call `trainer.scheduler_step()` to
    update the learning rate.
    6. Loop through each training step and call its callback function with the model, trainer, batch, and forward pass outputs.
    7. Return the `StepOutputs` object containing the forward pass outputs and loss values.
    """

    state = trainer.state
    forward: Union[tensor_dict_type, List[tensor_dict_type]] = {}
    loss_dict = {}
    update_fn = get_update_fn(trainer)
    any_update = False
    performed_scheduler_step = False
    # sanity check
    fw_has_grad = True
    fw_train_step: Any = ()
    for i, train_step in enumerate(train_steps):
        if i == 0 or train_step.requires_new_forward:
            fw_has_grad = train_step.requires_grad_in_forward
            fw_train_step = train_step
        if not fw_has_grad and train_step.requires_grad_in_forward:
            fw_name = fw_train_step.__class__.__name__
            current_name = train_step.__class__.__name__
            raise ValueError(
                f"current forward pass comes from '{fw_name}' and has no grad, "
                f"but '{current_name}' requires grad in forward. You can either set "
                f"`requires_grad_in_forward` of '{fw_name}' to True, or set "
                f"`requires_new_forward` of '{current_name}' to True."
            )
    # run train steps
    get_fw = lambda: m.run(batch_idx, batch, state, **forward_kwargs)
    for i, train_step in enumerate(train_steps):
        if train_step.should_skip(m, state):
            continue
        if i == 0 or train_step.requires_new_forward:
            with no_grad_context(enabled=not train_step.requires_grad_in_forward):
                if train_step.num_forward == 1:
                    forward = get_fw()
                else:
                    forward = [get_fw() for _ in range(train_step.num_forward)]
        optimizer = trainer.optimizers[train_step.scope]
        with toggle_optimizer(m, optimizer, enabled=train_step.enable_toggle_optimizer):
            with autocast(enabled=trainer.config.mixed_precision != PrecisionType.NO):
                loss_res = train_step.loss_fn(m, state, batch, forward, **loss_kwargs)
            update = state.step % train_step.grad_accumulate == 0
            update_fn(loss_res.loss, optimizer, update)
            loss_dict.update(loss_res.losses)
        if update:
            any_update = True
            performed_scheduler_step = train_step.requires_scheduler_step
            if performed_scheduler_step:
                trainer.scheduler_step()
    if any_update and not performed_scheduler_step:
        trainer.scheduler_step()
    # callbacks
    for train_step in train_steps:
        train_step.callback(m, trainer, batch, forward)
    return StepOutputs(forward, loss_dict)


def weighted_loss_score(config: TrainerConfig, loss_items: Dict[str, float]) -> float:
    if not config.loss_metrics_weights:
        loss = loss_items.get(LOSS_KEY)
        if loss is not None:
            return -loss
        return -sum(loss_items.values()) / len(loss_items)
    score = 0.0
    for k, w in config.loss_metrics_weights.items():
        v = loss_items.get(k)
        if v is None:
            continue
        score -= v * w
    return score


class CustomTrainStepLoss(NamedTuple):
    loss: Tensor
    losses: Dict[str, float]


class CustomTrainStep(ABC):
    def __init__(
        self,
        scope: str = "all",
        *,
        num_forward: int = 1,
        grad_accumulate: int = 1,
        requires_new_forward: bool = False,
        requires_grad_in_forward: bool = True,
        requires_scheduler_step: bool = False,
        enable_toggle_optimizer: bool = True,
    ) -> None:
        self.scope = scope
        self.num_forward = num_forward
        self.grad_accumulate = grad_accumulate
        self.requires_new_forward = requires_new_forward
        self.requires_grad_in_forward = requires_grad_in_forward
        self.requires_scheduler_step = requires_scheduler_step
        self.enable_toggle_optimizer = enable_toggle_optimizer

    # abstract

    @abstractmethod
    def loss_fn(
        self,
        m: "ModelWithCustomSteps",
        state: Optional[TrainerState],
        batch: tensor_dict_type,
        forward_results: Union[tensor_dict_type, List[tensor_dict_type]],
        **kwargs: Any,
    ) -> CustomTrainStepLoss:
        pass

    # optional callbacks

    def should_skip(self, m: "ModelWithCustomSteps", state: TrainerState) -> bool:
        return False

    def callback(
        self,
        m: "ModelWithCustomSteps",
        trainer: "ITrainer",
        batch: tensor_dict_type,
        forward_results: Union[tensor_dict_type, List[tensor_dict_type]],
    ) -> None:
        pass


class ModelWithCustomSteps(IDLModel, metaclass=ABCMeta):
    custom_train_step: bool = True
    custom_evaluate_step: bool = True
    custom_params_groups: bool = False
    custom_ddp_initialization: bool = False

    # abstract

    @property
    @abstractmethod
    def train_steps(self) -> List[CustomTrainStep]:
        pass

    @abstractmethod
    def evaluate(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional["TrainerState"],
        weighted_loss_score_fn: Callable[[Dict[str, float]], float],
        forward_kwargs: Dict[str, Any],
    ) -> MetricsOutputs:
        pass

    # inheritance

    def permute_trainer_config(self, trainer_config: "TrainerConfig") -> None:
        if trainer_config.optimizer_settings is None:
            opt_settings = {}
            for step in self.train_steps:
                scope = step.scope
                opt_settings[scope] = dict(
                    optimizer=trainer_config.optimizer_name or "adam",
                    scheduler=trainer_config.scheduler_name,
                    optimizer_config=trainer_config.optimizer_config,
                    scheduler_config=trainer_config.scheduler_config,
                )
            trainer_config.optimizer_settings = opt_settings

    # optional callback

    def params_groups(self, m: nn.Module) -> List[Dict[str, Any]]:
        return []

    # api

    def train_step(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        trainer: "ITrainer",
        forward_kwargs: Dict[str, Any],
        loss_kwargs: Dict[str, Any],
    ) -> StepOutputs:
        return run_train_steps(
            self,
            self.train_steps,
            batch_idx=batch_idx,
            batch=batch,
            trainer=trainer,
            forward_kwargs=forward_kwargs,
            loss_kwargs=loss_kwargs,
        )

    def evaluate_step(
        self,
        config: "TrainerConfig",
        loader: IDataLoader,
        portion: float,
        state: Optional["TrainerState"],
        forward_kwargs: Optional[Dict[str, Any]] = None,
    ) -> MetricsOutputs:
        loss_score_fn = lambda loss_items: weighted_loss_score(config, loss_items)
        device = get_device(self)
        tensor_batcher = TensorBatcher(loader, device)
        is_positive: Dict[str, bool] = {}
        metric_values: Dict[str, List[float]] = {}
        final_scores = []
        for i, batch in enumerate(tensor_batcher):
            if i / len(tensor_batcher) >= portion:
                break
            out = self.evaluate(i, batch, state, loss_score_fn, forward_kwargs or {})
            final_scores.append(out.final_score)
            for k, v in out.metric_values.items():
                metric_values.setdefault(k, []).append(v)
                is_positive[k] = out.is_positive[k]
        return MetricsOutputs(
            sum(final_scores) / len(final_scores),
            {k: sum(v) / len(v) for k, v in metric_values.items()},
            is_positive,
        )


__all__ = [
    "CustomTrainStepLoss",
    "CustomTrainStep",
    "ModelWithCustomSteps",
]



import torch

from torch import nn
from functools import partial

from .basic_blocks import ConvBlock
from ..ops import FeaturesConnector


class UNetEncoder(nn.Module):
    def __init__(
        self,
        depth,
        ch,
        norm_layer,
        batchnorm_from,
        max_channels,
        backbone_from,
        backbone_channels=None,
        backbone_mode="",
    ):
        super(UNetEncoder, self).__init__()
        self.depth = depth
        self.backbone_from = backbone_from
        self.block_channels = []
        backbone_channels = [] if backbone_channels is None else backbone_channels[::-1]
        relu = partial(nn.ReLU, inplace=True)

        in_channels = 4
        out_channels = ch

        self.block0 = UNetDownBlock(
            in_channels,
            out_channels,
            norm_layer=norm_layer if batchnorm_from == 0 else None,
            activation=relu,
            pool=True,
            padding=1,
        )
        self.block_channels.append(out_channels)
        in_channels, out_channels = out_channels, min(2 * out_channels, max_channels)
        self.block1 = UNetDownBlock(
            in_channels,
            out_channels,
            norm_layer=norm_layer if 0 <= batchnorm_from <= 1 else None,
            activation=relu,
            pool=True,
            padding=1,
        )
        self.block_channels.append(out_channels)

        self.blocks_connected = nn.ModuleDict()
        self.connectors = nn.ModuleDict()
        for block_i in range(2, depth):
            in_channels, out_channels = out_channels, min(
                2 * out_channels, max_channels
            )
            if 0 <= backbone_from <= block_i and len(backbone_channels):
                stage_channels = backbone_channels.pop()
                connector = FeaturesConnector(
                    backbone_mode, in_channels, stage_channels, in_channels
                )
                self.connectors[f"connector{block_i}"] = connector
                in_channels = connector.output_channels
            self.blocks_connected[f"block{block_i}"] = UNetDownBlock(
                in_channels,
                out_channels,
                norm_layer=norm_layer if 0 <= batchnorm_from <= block_i else None,
                activation=relu,
                padding=1,
                pool=block_i < depth - 1,
            )
            self.block_channels.append(out_channels)

    def forward(self, x, backbone_features):
        backbone_features = [] if backbone_features is None else backbone_features[::-1]
        outputs = []

        block_input = x
        output, block_input = self.block0(block_input)
        outputs.append(output)
        output, block_input = self.block1(block_input)
        outputs.append(output)

        for block_i in range(2, self.depth):
            block = self.blocks_connected[f"block{block_i}"]
            connector_name = f"connector{block_i}"
            if connector_name in self.connectors:
                stage_features = backbone_features.pop()
                connector = self.connectors[connector_name]
                block_input = connector(block_input, stage_features)
            output, block_input = block(block_input)
            outputs.append(output)

        return outputs[::-1]


class UNetDecoder(nn.Module):
    def __init__(
        self,
        depth,
        encoder_blocks_channels,
        norm_layer,
        attention_layer=None,
        attend_from=3,
        image_fusion=False,
    ):
        super(UNetDecoder, self).__init__()
        self.up_blocks = nn.ModuleList()
        self.image_fusion = image_fusion
        in_channels = encoder_blocks_channels.pop()
        out_channels = in_channels
        # Last encoder layer doesn't pool, so there're only (depth - 1) deconvs
        for d in range(depth - 1):
            out_channels = (
                encoder_blocks_channels.pop()
                if len(encoder_blocks_channels)
                else in_channels // 2
            )
            stage_attention_layer = attention_layer if 0 <= attend_from <= d else None
            self.up_blocks.append(
                UNetUpBlock(
                    in_channels,
                    out_channels,
                    out_channels,
                    norm_layer=norm_layer,
                    activation=partial(nn.ReLU, inplace=True),
                    padding=1,
                    attention_layer=stage_attention_layer,
                )
            )
            in_channels = out_channels

        if self.image_fusion:
            self.conv_attention = nn.Conv2d(out_channels, 1, kernel_size=1)
        self.to_rgb = nn.Conv2d(out_channels, 3, kernel_size=1)

    def forward(self, encoder_outputs, input_image, mask):
        output = encoder_outputs[0]
        for block, skip_output in zip(self.up_blocks, encoder_outputs[1:]):
            output = block(output, skip_output, mask)

        if self.image_fusion:
            attention_map = torch.sigmoid(3.0 * self.conv_attention(output))
            output = attention_map * input_image + (1.0 - attention_map) * self.to_rgb(
                output
            )
        else:
            output = self.to_rgb(output)

        return output


class UNetDownBlock(nn.Module):
    def __init__(
        self, in_channels, out_channels, norm_layer, activation, pool, padding
    ):
        super(UNetDownBlock, self).__init__()
        self.convs = UNetDoubleConv(
            in_channels,
            out_channels,
            norm_layer=norm_layer,
            activation=activation,
            padding=padding,
        )
        self.pooling = nn.MaxPool2d(2, 2) if pool else nn.Identity()

    def forward(self, x):
        conv_x = self.convs(x)
        return conv_x, self.pooling(conv_x)


class UNetUpBlock(nn.Module):
    def __init__(
        self,
        in_channels_decoder,
        in_channels_encoder,
        out_channels,
        norm_layer,
        activation,
        padding,
        attention_layer,
    ):
        super(UNetUpBlock, self).__init__()
        self.upconv = nn.Sequential(
            nn.Upsample(scale_factor=2, mode="bilinear", align_corners=True),
            ConvBlock(
                in_channels_decoder,
                out_channels,
                kernel_size=3,
                stride=1,
                padding=1,
                norm_layer=None,
                activation=activation,
            ),
        )
        self.convs = UNetDoubleConv(
            in_channels_encoder + out_channels,
            out_channels,
            norm_layer=norm_layer,
            activation=activation,
            padding=padding,
        )
        if attention_layer is not None:
            self.attention = attention_layer(
                in_channels_encoder + out_channels, norm_layer, activation
            )
        else:
            self.attention = None

    def forward(self, x, encoder_out, mask=None):
        upsample_x = self.upconv(x)
        x_cat_encoder = torch.cat([encoder_out, upsample_x], dim=1)
        if self.attention is not None:
            x_cat_encoder = self.attention(x_cat_encoder, mask)
        return self.convs(x_cat_encoder)


class UNetDoubleConv(nn.Module):
    def __init__(self, in_channels, out_channels, norm_layer, activation, padding):
        super(UNetDoubleConv, self).__init__()
        self.block = nn.Sequential(
            ConvBlock(
                in_channels,
                out_channels,
                kernel_size=3,
                stride=1,
                padding=padding,
                norm_layer=norm_layer,
                activation=activation,
            ),
            ConvBlock(
                out_channels,
                out_channels,
                kernel_size=3,
                stride=1,
                padding=padding,
                norm_layer=norm_layer,
                activation=activation,
            ),
        )

    def forward(self, x):
        return self.block(x)



import numpy as np

from typing import Union
from PIL.Image import Image
from cftool.cv import read_image

from .inference.utils import load_model
from .inference.predictor import Predictor
from .....misc.toolkit import download_model


class ImageHarmonizationAPI:
    def __init__(self, device: str = "cpu", *, use_half: bool = False) -> None:
        m = load_model("hrnet32_idih256", download_model("hrnet"))
        self.device = device
        self.predictor = Predictor(m, device, with_flip=False)

    def to(self, device: str, *, use_half: bool = False) -> None:
        self.device = device
        self.predictor.to(device)

    # return uint8 image array
    def predict(self, image: np.ndarray, mask: np.ndarray) -> np.ndarray:
        return self.predictor.predict(image, mask)

    # return uint8 image array
    def run(self, image: Union[str, Image], mask: Union[str, Image]) -> np.ndarray:
        image_arr = read_image(
            image,
            None,
            anchor=None,
            normalize=False,
            to_torch_fmt=False,
        ).image
        mask_arr = read_image(
            mask,
            None,
            anchor=None,
            to_mask=True,
            to_torch_fmt=False,
        ).image
        return self.predict(image_arr, mask_arr)


__all__ = [
    "ImageHarmonizationAPI",
]



import torch
import torch.nn as nn

from ..modeling.conv_autoencoder import ConvEncoder
from ..modeling.conv_autoencoder import DeconvDecoder


class DeepImageHarmonization(nn.Module):
    def __init__(
        self,
        depth,
        norm_layer=nn.BatchNorm2d,
        batchnorm_from=0,
        attend_from=-1,
        image_fusion=False,
        ch=64,
        max_channels=512,
        backbone_from=-1,
        backbone_channels=None,
        backbone_mode="",
    ):
        super(DeepImageHarmonization, self).__init__()
        self.depth = depth
        self.encoder = ConvEncoder(
            depth,
            ch,
            norm_layer,
            batchnorm_from,
            max_channels,
            backbone_from,
            backbone_channels,
            backbone_mode,
        )
        self.decoder = DeconvDecoder(
            depth, self.encoder.blocks_channels, norm_layer, attend_from, image_fusion
        )

    def forward(self, image, mask, backbone_features=None):
        x = torch.cat((image, mask), dim=1)
        intermediates = self.encoder(x, backbone_features)
        output = self.decoder(intermediates, image, mask)
        return {"images": output}



import numpy as np

from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from dataclasses import dataclass
from cftool.misc import safe_execute
from cftool.misc import shallow_copy_dict
from cftool.array import normalize
from cftool.array import normalize_from
from cftool.array import recover_normalize_from
from cftool.array import min_max_normalize
from cftool.array import min_max_normalize_from
from cftool.array import recover_min_max_normalize_from
from cftool.array import quantile_normalize
from cftool.array import quantile_normalize_from
from cftool.array import recover_quantile_normalize_from

from .recognizer import RecognizerBlock
from ....schema import DataBundle
from ....schema import ColumnTypes
from ....schema import INoInitDataBlock


class PreProcessMethod(str, Enum):
    MIN_MAX = "min_max"
    NORMALIZE = "normalize"
    QUANTILE_NORMALIZE = "quantile_normalize"


fn_mapping = {
    PreProcessMethod.MIN_MAX: min_max_normalize,
    PreProcessMethod.NORMALIZE: normalize,
    PreProcessMethod.QUANTILE_NORMALIZE: quantile_normalize,
}
fn_with_stats_mapping = {
    PreProcessMethod.MIN_MAX: min_max_normalize_from,
    PreProcessMethod.NORMALIZE: normalize_from,
    PreProcessMethod.QUANTILE_NORMALIZE: quantile_normalize_from,
}
recover_fn_mapping = {
    PreProcessMethod.MIN_MAX: recover_min_max_normalize_from,
    PreProcessMethod.NORMALIZE: recover_normalize_from,
    PreProcessMethod.QUANTILE_NORMALIZE: recover_quantile_normalize_from,
}


@dataclass
class MLPreProcessConfig:
    auto_preprocess: bool = True
    preprocess_methods: Optional[Dict[str, PreProcessMethod]] = None
    preprocess_configs: Optional[Dict[str, Dict[str, Any]]] = None
    label_preprocess_methods: Optional[Dict[str, PreProcessMethod]] = None
    label_preprocess_configs: Optional[Dict[str, Dict[str, Any]]] = None


def _fit_transform(
    data: np.ndarray,
    target: List[int],
    auto_preprocess: bool,
    preprocess_methods: Dict[str, PreProcessMethod],
    preprocess_configs: Dict[str, Dict[str, Any]],
    methods: Dict[str, PreProcessMethod],
    all_stats: Dict[str, Dict[str, float]],
) -> None:
    for idx in target:
        str_idx = str(idx)
        method = preprocess_methods.get(str_idx)
        if method is None and not auto_preprocess:
            continue
        if method is None:
            method = PreProcessMethod.NORMALIZE
        fn = fn_mapping.get(method)
        if fn is None:
            raise ValueError(f"unrecognized method '{method}' occurred")
        kw = shallow_copy_dict(preprocess_configs.get(str_idx, {}))
        kw["arr"] = data[..., idx]
        kw["return_stats"] = True
        array, stats = safe_execute(fn, kw)
        data[..., idx] = array
        methods[str_idx] = method
        all_stats[str_idx] = stats


def _transform(
    data: np.ndarray,
    methods: Dict[str, PreProcessMethod],
    all_stats: Dict[str, Dict[str, float]],
    mapping: Dict[PreProcessMethod, Any],
) -> None:
    for str_idx, stats in all_stats.items():
        idx = int(str_idx)
        fn = mapping[methods[str_idx]]
        data[..., idx] = fn(data[..., idx], stats)


@INoInitDataBlock.register("ml_preprocessor")
class PreProcessorBlock(INoInitDataBlock):
    config: MLPreProcessConfig  # type: ignore
    methods: Dict[str, PreProcessMethod]
    stats: Dict[str, Dict[str, float]]
    label_methods: Dict[str, PreProcessMethod]
    label_stats: Dict[str, Dict[str, float]]

    # inheritance

    def to_info(self) -> Dict[str, Any]:
        return dict(
            methods=self.methods,
            stats=self.stats,
            label_methods=self.label_methods,
            label_stats=self.label_stats,
        )

    def transform(self, bundle: DataBundle, for_inference: bool) -> DataBundle:
        self._transform_x(bundle.x_train)
        if bundle.y_train is not None:
            self._transform_y(bundle.y_train)
        if bundle.x_valid is not None:
            self._transform_x(bundle.x_valid)
        if bundle.y_valid is not None:
            self._transform_y(bundle.y_valid)
        return bundle

    def fit_transform(self, bundle: DataBundle) -> DataBundle:
        x_train = bundle.x_train
        y_train = bundle.y_train
        if not isinstance(x_train, np.ndarray):
            raise ValueError("`PreProcessorBlock` should be used on numpy features.")
        if not isinstance(y_train, np.ndarray):
            raise ValueError("`PreProcessorBlock` should be used on numpy labels.")
        recognizer = self.recognizer
        if recognizer is None:
            x_target = list(range(x_train.shape[-1]))
            y_target = list(range(y_train.shape[-1]))
        else:
            mapping = recognizer.index_mapping
            x_target = [
                mapping[str_idx]
                for str_idx, column_type in recognizer.feature_types.items()
                if column_type == ColumnTypes.NUMERICAL
            ]
            y_target = [
                int(str_idx)
                for str_idx, column_type in recognizer.label_types.items()
                if column_type == ColumnTypes.NUMERICAL
            ]
        if self.config.preprocess_methods is None:
            self.config.preprocess_methods = {}
        if self.config.preprocess_configs is None:
            self.config.preprocess_configs = {}
        if self.config.label_preprocess_methods is None:
            self.config.label_preprocess_methods = {}
        if self.config.label_preprocess_configs is None:
            self.config.label_preprocess_configs = {}
        self.methods = {}
        self.stats = {}
        self.label_methods = {}
        self.label_stats = {}
        _fit_transform(
            x_train,
            x_target,
            self.config.auto_preprocess,
            self.config.preprocess_methods,
            self.config.preprocess_configs,
            self.methods,
            self.stats,
        )
        _fit_transform(
            y_train,
            y_target,
            self.config.auto_preprocess,
            self.config.label_preprocess_methods,
            self.config.label_preprocess_configs,
            self.label_methods,
            self.label_stats,
        )
        if bundle.x_valid is not None:
            self._transform_x(bundle.x_valid)
        if bundle.y_valid is not None:
            self._transform_y(bundle.y_valid)
        return bundle

    def recover_labels(self, y: np.ndarray) -> np.ndarray:
        _transform(y, self.label_methods, self.label_stats, recover_fn_mapping)
        return y

    # api

    @property
    def recognizer(self) -> Optional[RecognizerBlock]:
        return self.try_get_previous(RecognizerBlock.__identifier__)

    # internal

    def _transform_x(self, x: np.ndarray) -> None:
        _transform(x, self.methods, self.stats, fn_with_stats_mapping)

    def _transform_y(self, y: np.ndarray) -> None:
        _transform(y, self.label_methods, self.label_stats, fn_with_stats_mapping)


__all__ = [
    "PreProcessMethod",
    "MLPreProcessConfig",
    "PreProcessorBlock",
]



import os

from cftool.cv import save_images
from cftool.array import to_device

from .general import ImageCallback
from ..misc.toolkit import eval_context
from ..misc.toolkit import make_indices_visualization_map
from ..schema import ITrainer
from ..constants import INPUT_KEY
from ..constants import PREDICTIONS_KEY


@ImageCallback.register("clf")
class ClassificationCallback(ImageCallback):
    def log_artifacts(self, trainer: ITrainer) -> None:
        if not self.is_local_rank_0:
            return None
        batch = next(iter(trainer.validation_loader))
        batch = to_device(batch, trainer.device)
        original = batch[INPUT_KEY]
        with eval_context(trainer.model):
            logits = trainer.model.classify(original)[PREDICTIONS_KEY]
            labels_map = make_indices_visualization_map(logits.argmax(1))
        image_folder = self._prepare_folder(trainer)
        save_images(original, os.path.join(image_folder, "original.png"))
        save_images(labels_map, os.path.join(image_folder, "labels.png"))


__all__ = [
    "ClassificationCallback",
]



import torch

from PIL import Image
from cftool.cv import to_rgb

try:
    from lavis.models import load_model_and_preprocess
except:
    load_model_and_preprocess = None


class BLIPAPI:
    def __init__(self, device: str = "cpu", *, use_half: bool = False) -> None:
        if load_model_and_preprocess is None:
            raise ValueError("`salesforce-lavis` is needed for `BLIPAPI`")
        kw = dict(
            name="blip_caption",
            model_type="base_coco",
            is_eval=True,
            device=device,
        )
        self.device = device
        self.model, self.processors, _ = load_model_and_preprocess(**kw)

    def to(self, device: str, *, use_half: bool = False) -> None:
        self.device = device
        self.model.to(device)

    @torch.no_grad()
    def caption(self, image: Image.Image) -> str:
        rgb_image = to_rgb(image)
        net = self.processors["eval"](rgb_image)[None].to(self.device)
        return self.model.generate(dict(image=net))[0]


__all__ = [
    "BLIPAPI",
]



import torch

import numpy as np

from abc import abstractmethod
from enum import Enum
from numpy import ndarray
from typing import Any
from typing import List
from typing import Tuple
from typing import Union
from typing import Optional
from typing import NamedTuple
from PIL.Image import Image
from cftool.cv import to_uint8
from cftool.cv import read_image

from ....misc.toolkit import download_model

try:
    import cv2
except:
    cv2 = None


def ceil_mod(x: int, mod: int) -> int:
    if x % mod == 0:
        return x
    return (x // mod + 1) * mod


def pad_img_to_mod(
    img: ndarray,
    mod: int,
    square: bool = False,
    min_size: Optional[int] = None,
) -> np.array:
    if len(img.shape) == 2:
        img = img[..., None]
    height, width = img.shape[:2]
    out_height = ceil_mod(height, mod)
    out_width = ceil_mod(width, mod)

    if min_size is not None:
        assert min_size % mod == 0
        out_width = max(min_size, out_width)
        out_height = max(min_size, out_height)

    if square:
        max_size = max(out_height, out_width)
        out_height = max_size
        out_width = max_size

    return np.pad(
        img,
        ((0, out_height - height), (0, out_width - width), (0, 0)),
        mode="symmetric",
    )


def boxes_from_mask(mask: ndarray) -> List[ndarray]:
    height, width = mask.shape[:2]
    _, thresh = cv2.threshold(mask, 127, 255, 0)
    contours, _ = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    boxes = []
    for cnt in contours:
        x, y, w, h = cv2.boundingRect(cnt)
        box = np.array([x, y, x + w, y + h]).astype(int)
        box[::2] = np.clip(box[::2], 0, width)
        box[1::2] = np.clip(box[1::2], 0, height)
        boxes.append(box)

    return boxes


def resize_max_size(
    np_img: ndarray,
    size_limit: int,
    interpolation: int = 0 if cv2 is None else cv2.INTER_CUBIC,
) -> ndarray:
    # Resize image's longer size to size_limit if longer size larger than size_limit
    h, w = np_img.shape[:2]
    if max(h, w) > size_limit:
        ratio = size_limit / max(h, w)
        new_w = int(w * ratio + 0.5)
        new_h = int(h * ratio + 0.5)
        return cv2.resize(np_img, dsize=(new_w, new_h), interpolation=interpolation)
    else:
        return np_img


def to_tensor(np_img: ndarray) -> torch.Tensor:
    net = np_img.transpose([2, 0, 1])[None]
    net = torch.from_numpy(net)
    return net


class HDStrategy(str, Enum):
    ORIGINAL = "original"
    RESIZE = "resize"
    CROP = "crop"


class Config(NamedTuple):
    hd_strategy: HDStrategy = HDStrategy.CROP
    hd_strategy_resize_limit: int = 2048
    hd_strategy_crop_trigger_size: int = 1280
    hd_strategy_crop_margin: int = 196


class InpaintModel:
    min_size: Optional[int] = None
    pad_mod: int = 8
    pad_to_square: bool = False

    def __init__(self, device: str = "cpu", *, use_half: bool = False):
        self.device = device
        self.init_model(device, use_half=use_half)

    @abstractmethod
    def init_model(self, device: str, *, use_half: bool = False) -> None:
        pass

    @abstractmethod
    def forward(
        self,
        image: ndarray,
        mask: ndarray,
        config: Config,
    ) -> ndarray:
        pass

    def _pad_and_run(self, image: ndarray, mask: ndarray, config: Config) -> ndarray:
        origin_height, origin_width = image.shape[:2]
        pad_image = pad_img_to_mod(
            image,
            mod=self.pad_mod,
            square=self.pad_to_square,
            min_size=self.min_size,
        )
        pad_mask = pad_img_to_mod(
            mask,
            mod=self.pad_mod,
            square=self.pad_to_square,
            min_size=self.min_size,
        )

        result = self.forward(pad_image, pad_mask, config)
        result = result[:origin_height, :origin_width, :]

        result, image, mask = self.forward_post_process(result, image, mask, config)

        mask = mask[..., None]
        result = result * mask + image * (1.0 - mask)
        return result

    def forward_post_process(
        self,
        result: ndarray,
        image: ndarray,
        mask: ndarray,
        config: Config,
    ) -> Tuple[ndarray, ndarray, ndarray]:
        return result, image, mask

    @torch.no_grad()
    def __call__(self, image: ndarray, mask: ndarray, config: Config) -> ndarray:
        size_limit = config.hd_strategy_crop_trigger_size
        if max(image.shape) <= size_limit:
            result = self._pad_and_run(image, mask, config)
        else:
            if config.hd_strategy == HDStrategy.CROP:
                boxes = boxes_from_mask(to_uint8(mask))
                crop_result = []
                for box in boxes:
                    crop_image, crop_box = self._run_box(image, mask, box, config)
                    crop_result.append((crop_image, crop_box))
                result = image
                for crop_image, crop_box in crop_result:
                    x1, y1, x2, y2 = crop_box
                    result[y1:y2, x1:x2, :] = crop_image
            elif config.hd_strategy == HDStrategy.RESIZE:
                origin_size = image.shape[:2]
                downsize_image = resize_max_size(image, size_limit=size_limit)
                downsize_mask = resize_max_size(mask, size_limit=size_limit)
                result = self._pad_and_run(downsize_image, downsize_mask, config)
                result = cv2.resize(
                    result,
                    (origin_size[1], origin_size[0]),
                    interpolation=cv2.INTER_CUBIC,
                )
                original_pixel_indices = mask < 127
                result[original_pixel_indices] = image[original_pixel_indices]
            else:
                msg = f"unrecognized hd_strategy `{config.hd_strategy}` occurred"
                raise ValueError(msg)
        torch.cuda.empty_cache()
        return result

    def _crop_box(
        self,
        image: ndarray,
        mask: ndarray,
        box: ndarray,
        config: Config,
    ) -> Tuple[ndarray, ndarray, List[int]]:
        box_h = box[3] - box[1]
        box_w = box[2] - box[0]
        cx = (box[0] + box[2]) // 2
        cy = (box[1] + box[3]) // 2
        img_h, img_w = image.shape[:2]

        w = box_w + config.hd_strategy_crop_margin * 2
        h = box_h + config.hd_strategy_crop_margin * 2

        _l = cx - w // 2
        _r = cx + w // 2
        _t = cy - h // 2
        _b = cy + h // 2

        l = max(_l, 0)
        r = min(_r, img_w)
        t = max(_t, 0)
        b = min(_b, img_h)

        if _l < 0:
            r += abs(_l)
        if _r > img_w:
            l -= _r - img_w
        if _t < 0:
            b += abs(_t)
        if _b > img_h:
            t -= _b - img_h

        l = max(l, 0)
        r = min(r, img_w)
        t = max(t, 0)
        b = min(b, img_h)

        crop_img = image[t:b, l:r, :]
        crop_mask = mask[t:b, l:r]

        return crop_img, crop_mask, [l, t, r, b]

    def _run_box(
        self,
        image: ndarray,
        mask: ndarray,
        box: ndarray,
        config: Config,
    ) -> Tuple[ndarray, List[int]]:
        crop_img, crop_mask, [l, t, r, b] = self._crop_box(image, mask, box, config)
        return self._pad_and_run(crop_img, crop_mask, config), [l, t, r, b]


class LaMa(InpaintModel):
    pad_mod = 8

    def init_model(self, device: str, *, use_half: bool = False) -> None:
        if cv2 is None:
            raise ValueError("`cv2` is needed for `LaMa`")
        model_path = download_model("lama")
        model = torch.jit.load(model_path, map_location="cpu")
        model = model.to(device)
        model.eval()
        self.model = model
        self.model_path = model_path

    def to(self, device: str, *, use_half: bool = False) -> None:
        self.device = device
        self.model.to(device)

    def forward(self, image: ndarray, mask: ndarray, config: Config) -> ndarray:
        image, mask = map(to_tensor, [image, mask])
        image = image.to(self.device)
        mask = mask.to(self.device)
        net = self.model(image, mask)
        net = net[0].permute(1, 2, 0).cpu().numpy()
        return net


class LaMaAPI:
    def __init__(self, device: torch.device) -> None:
        self.lama = LaMa(device)

    def inpaint(self, image: Union[str, Image], mask: Union[str, Image]) -> np.ndarray:
        cfg = Config()
        image_arr = read_image(
            image,
            None,
            anchor=None,
            to_torch_fmt=False,
        ).image
        mask_arr = read_image(
            mask,
            None,
            anchor=None,
            to_mask=True,
            to_torch_fmt=False,
        ).image
        mask_arr[mask_arr > 0.0] = 1.0
        return self.lama(image_arr, mask_arr, cfg)


__all__ = [
    "LaMaAPI",
]



import torch

from torch import nn
from torch import Tensor

from .schema import ChannelMixerBase
from ..convs import DepthWiseConv2d
from ..common import Lambda
from ..hijacks import HijackCustomLinear
from ..activations import GEGLU
from ..activations import Activation


@ChannelMixerBase.register("ff")
class FeedForward(ChannelMixerBase):
    def __init__(
        self,
        in_dim: int,
        latent_dim: int,
        dropout: float,
        activation: str = "GELU",
        add_last_dropout: bool = True,
    ):
        super().__init__(in_dim, latent_dim, dropout)
        if activation == "geglu":
            blocks = [GEGLU(in_dim, latent_dim)]
        else:
            blocks = [
                HijackCustomLinear(in_dim, latent_dim),
                Activation.make(activation),
            ]
        blocks += [nn.Dropout(dropout), HijackCustomLinear(latent_dim, in_dim)]
        if add_last_dropout:
            blocks.append(nn.Dropout(dropout))
        self.net = nn.Sequential(*blocks)

    @property
    def need_2d(self) -> bool:
        return False

    def forward(self, net: Tensor) -> Tensor:
        return self.net(net)


@ChannelMixerBase.register("mix_ff")
class MixFeedForward(ChannelMixerBase):
    def __init__(self, in_dim: int, latent_dim: int, dropout: float):
        super().__init__(in_dim, latent_dim, dropout)
        self.net = nn.Sequential(
            HijackCustomLinear(in_dim, latent_dim),
            Lambda(lambda t: t.permute(0, 3, 1, 2), "permute -> BCHW"),
            DepthWiseConv2d(latent_dim),
            Lambda(lambda t: t.flatten(2).transpose(1, 2), "transpose -> BNC"),
            nn.GELU(),
            nn.Dropout(dropout),
            HijackCustomLinear(latent_dim, in_dim),
            nn.Dropout(dropout),
        )

    @property
    def need_2d(self) -> bool:
        return True

    def forward(self, net: Tensor) -> Tensor:
        return self.net(net)


@ChannelMixerBase.register("rwkv")
class RWKVChannelMixer(ChannelMixerBase):
    def __init__(
        self,
        layer_idx: int,
        num_layers: int,
        in_dim: int,
        latent_dim: int,
        dropout: float,
    ):
        super().__init__(in_dim, latent_dim, dropout)
        self.time_shift = nn.ZeroPad2d((0, 0, 1, -1))

        self.key = nn.Linear(in_dim, latent_dim, bias=False)
        self.value = nn.Linear(latent_dim, in_dim, bias=False)
        self.receptance = nn.Linear(in_dim, in_dim, bias=False)

        with torch.no_grad():
            for k, v in self.state_dict().items():
                for zero_k in ["value", "receptance"]:
                    if zero_k in k:
                        nn.init.zeros_(v)
                        break
                else:
                    nn.init.orthogonal_(v)

            _1_to_almost0 = 1.0 - (layer_idx / num_layers)
            x = torch.ones(1, 1, in_dim)
            for i in range(in_dim):
                x[0, 0, i] = i / in_dim
            self.time_mix_k = nn.Parameter(torch.pow(x, _1_to_almost0))
            self.time_mix_r = nn.Parameter(torch.pow(x, _1_to_almost0))

    @property
    def need_2d(self) -> bool:
        return False

    def forward(self, net: Tensor) -> Tensor:
        shifted = self.time_shift(net)
        net_k = net * self.time_mix_k + shifted * (1.0 - self.time_mix_k)
        net_r = net * self.time_mix_r + shifted * (1.0 - self.time_mix_r)
        net_k = self.key(net_k)
        net_k = torch.square(torch.relu(net_k))
        net_kv = self.value(net_k)
        return torch.sigmoid(self.receptance(net_r)) * net_kv


__all__ = [
    "FeedForward",
    "MixFeedForward",
    "RWKVChannelMixer",
]



import numpy as np

from enum import Enum
from typing import Optional
from typing import NamedTuple
from cftool.types import np_dict_type


class MLDatasetTag(str, Enum):
    TRAIN = "train"
    VALID = "validation"


class MLBatch(NamedTuple):
    input: np.ndarray
    labels: Optional[np.ndarray]
    others: Optional[np_dict_type] = None


__all__ = [
    "MLDatasetTag",
    "MLBatch",
]



import numpy as np

from typing import List
from typing import Tuple
from cftool.types import np_dict_type

from ....schema import IRuntimeDataBlock
from ....constants import INPUT_KEY


@IRuntimeDataBlock.register("static_normalize")
class StaticNormalizeBlock(IRuntimeDataBlock):
    division: float

    def __init__(self, division: float = 255.0) -> None:
        super().__init__(division=division)

    @property
    def fields(self) -> List[str]:
        return ["division"]

    def postprocess_item(self, item: np_dict_type) -> np_dict_type:
        item[INPUT_KEY] = item[INPUT_KEY].astype(np.float64) / self.division
        return item


@IRuntimeDataBlock.register("affine_normalize")
class AffineNormalizeBlock(IRuntimeDataBlock):
    center: float
    scale: float

    def __init__(self, center: float = 0.5, scale: float = 0.5) -> None:
        super().__init__(center=center, scale=scale)

    @property
    def fields(self) -> List[str]:
        return ["center", "scale"]

    def postprocess_item(self, item: np_dict_type) -> np_dict_type:
        inp = item[INPUT_KEY].astype(np.float64)
        inp = (inp - self.center) / self.scale
        item[INPUT_KEY] = inp
        return item


@IRuntimeDataBlock.register("imagenet_normalize")
class ImagenetNormalizeBlock(IRuntimeDataBlock):
    mean: List[float]
    std: List[float]

    def __init__(
        self,
        mean: Tuple[float, float, float] = (0.485, 0.456, 0.406),
        std: Tuple[float, float, float] = (0.229, 0.224, 0.225),
    ) -> None:
        super().__init__(mean=list(mean), std=list(std))

    @property
    def fields(self) -> List[str]:
        return ["mean", "std"]

    def postprocess_item(self, item: np_dict_type) -> np_dict_type:
        inp = item[INPUT_KEY].astype(np.float64)
        inp = (inp - self.mean) / self.std
        item[INPUT_KEY] = inp
        return item


__all__ = [
    "StaticNormalizeBlock",
    "AffineNormalizeBlock",
    "ImagenetNormalizeBlock",
]



import torch

import numpy as np

from typing import List
from typing import Tuple
from typing import Union
from torchvision import transforms
from cftool.types import np_dict_type

from ....schema import IRuntimeDataBlock
from ....constants import INPUT_KEY


@IRuntimeDataBlock.register("center_crop")
class CenterCropBlock(IRuntimeDataBlock):
    size: Union[int, List[int]]

    def __init__(self, size: Union[int, List[int]] = 512) -> None:
        super().__init__(size=size)

    @property
    def fields(self) -> List[str]:
        return ["size"]

    def postprocess_item(self, item: np_dict_type) -> np_dict_type:
        tensor = torch.from_numpy(item[INPUT_KEY])
        tensor = transforms.CenterCrop(self.size)(tensor)
        item[INPUT_KEY] = tensor.numpy()
        return item


@IRuntimeDataBlock.register("random_crop")
class RandomCropBlock(IRuntimeDataBlock):
    size: Union[int, List[int]]

    def __init__(self, size: Union[int, List[int]] = 512) -> None:
        super().__init__(size=size)

    @property
    def fields(self) -> List[str]:
        return ["size"]

    def postprocess_item(self, item: np_dict_type) -> np_dict_type:
        tensor = torch.from_numpy(item[INPUT_KEY])
        tensor = transforms.RandomCrop(self.size)(tensor)
        item[INPUT_KEY] = tensor.numpy()
        return item


__all__ = [
    "CenterCropBlock",
    "RandomCropBlock",
]



from typing import Any
from typing import Dict
from typing import Optional

from .schema import Encoder1DFromPatches
from ....modules.blocks import MixedStackedEncoder


@Encoder1DFromPatches.register("mixer")
class MixerEncoder(Encoder1DFromPatches):
    def __init__(
        self,
        *,
        img_size: int,
        patch_size: int,
        in_channels: int,
        latent_dim: int = 128,
        to_patches_type: str = "vanilla",
        to_patches_config: Optional[Dict[str, Any]] = None,
        num_layers: int = 4,
        norm_type: Optional[str] = "layer",
    ):
        super().__init__(
            img_size=img_size,
            patch_size=patch_size,
            in_channels=in_channels,
            latent_dim=latent_dim,
            to_patches_type=to_patches_type,
            to_patches_config=to_patches_config,
        )
        self.encoder = MixedStackedEncoder(
            latent_dim,
            self.num_patches,
            token_mixing_type="mlp",
            num_layers=num_layers,
            norm_type=norm_type,
        )


__all__ = ["MixerEncoder"]



import os
import re
import json
import math
import torch

import torch.distributed as dist

from typing import Any
from typing import Set
from typing import Dict
from typing import List
from typing import Tuple
from typing import Callable
from typing import Optional
from accelerate import Accelerator
from tqdm.autonotebook import tqdm
from torch.optim import Optimizer
from cftool.misc import print_info
from cftool.misc import print_error
from cftool.misc import safe_execute
from cftool.misc import print_warning
from cftool.misc import shallow_copy_dict
from cftool.misc import sort_dict_by_value
from cftool.array import to_device
from cftool.types import tensor_dict_type
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data.distributed import DistributedSampler

from .schema import ITrainer
from .schema import StepOutputs
from .schema import ILoss
from .schema import TqdmSettings
from .schema import TrainerState
from .schema import IDLModel
from .schema import MetricsOutputs
from .schema import MonitorResults
from .schema import TrainerMonitor
from .schema import IMetric
from .schema import MultipleMetrics
from .schema import TrainerCallback
from .schema import IInference
from .schema import IDataLoader
from .schema import IData
from .schema import PrecisionType
from .schema import TrainerConfig
from .constants import LOSS_KEY
from .constants import PT_PREFIX
from .constants import SCORES_FILE
from .constants import CHECKPOINTS_FOLDER
from .data.utils import TensorBatcher
from .data.pytorch import TorchDataLoader
from .misc.toolkit import summary
from .misc.toolkit import get_ddp_info
from .misc.toolkit import eval_context
from .models.schemas import ModelWithCustomSteps
from .modules.schedulers import WarmupScheduler


class amp_context:
    def __init__(self, enabled: bool):
        if not enabled:
            self.ctx = None
        else:
            self.ctx = torch.cuda.amp.autocast(enabled=True)

    def __enter__(self) -> None:
        if self.ctx is not None:
            self.ctx.__enter__()

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if self.ctx is not None:
            self.ctx.__exit__(exc_type, exc_val, exc_tb)


def get_scores(folder: str) -> Dict[str, float]:
    scores_path = os.path.join(folder, SCORES_FILE)
    if not os.path.isfile(scores_path):
        return {}
    with open(scores_path, "r") as f:
        return json.load(f)


def get_sorted_checkpoints(checkpoint_folder: str) -> List[str]:
    # better checkpoints will be placed earlier,
    #  which means `checkpoints[0]` is the best checkpoint
    scores = get_scores(checkpoint_folder)
    if not scores:
        return []
    return list(sort_dict_by_value(scores, reverse=True).keys())


def weighted_loss_score(config: TrainerConfig, loss_items: Dict[str, float]) -> float:
    if not config.loss_metrics_weights:
        loss = loss_items.get(LOSS_KEY)
        if loss is not None:
            return -loss
        return -sum(loss_items.values()) / len(loss_items)
    score = 0.0
    for k, w in config.loss_metrics_weights.items():
        v = loss_items.get(k)
        if v is None:
            continue
        score -= v * w
    return score


def get_metrics(
    config: TrainerConfig,
    model: IDLModel,
    loss: ILoss,
    metrics: Optional[IMetric],
    inference: IInference,
    loader: IDataLoader,
    *,
    portion: float = 1.0,
    state: Optional[TrainerState] = None,
    forward_kwargs: Optional[Dict[str, Any]] = None,
) -> MetricsOutputs:
    if isinstance(model, ModelWithCustomSteps) and model.custom_evaluate_step:
        use_grad = inference.use_grad_in_predict
        args = config, loader, portion, state, forward_kwargs
        try:
            with eval_context(model, use_grad=use_grad):
                rs = model.evaluate_step(*args)
        except:
            inference.use_grad_in_predict = True
            with eval_context(model, use_grad=True):
                rs = model.evaluate_step(*args)
        return rs
    outputs = inference.get_outputs(
        loader,
        portion=portion,
        state=state,
        metrics=metrics,
        loss=loss if config.use_losses_as_metrics else None,
        return_outputs=False,
        **(forward_kwargs or {}),
    )
    metric_values = {}
    is_positive = {}
    final_scores = []
    loss_items = outputs.loss_items
    metric_outputs = outputs.metric_outputs
    if loss_items is not None:
        metric_values.update(loss_items)
        is_positive.update({k: False for k in loss_items})
        final_scores.append(weighted_loss_score(config, loss_items))
    if metric_outputs is not None:
        metric_values.update(metric_outputs.metric_values)
        is_positive.update(metric_outputs.is_positive)
        final_scores.append(metric_outputs.final_score)
    final_score = sum(final_scores) / len(final_scores)
    return MetricsOutputs(final_score, metric_values, is_positive)


def get_input_sample(loader: IDataLoader, device: torch.device) -> tensor_dict_type:
    sample = next(iter(TensorBatcher(loader, device)))
    for k, v in sample.items():
        if isinstance(v, torch.Tensor):
            sample[k] = v[:1]
        elif isinstance(v, list):
            sample[k] = [vv[:1] if isinstance(vv, torch.Tensor) else vv for vv in v]
        else:
            sample[k] = v
    return sample


class Trainer(ITrainer):
    model_log_file = "model.txt"
    metrics_log_file = "metrics.txt"
    summary_log_file = "summary.txt"

    def __init__(self, config: TrainerConfig):
        self.config = config
        self.tqdm_settings = safe_execute(TqdmSettings, config.tqdm_settings or {})
        self._current_scheduler_epoch = -1
        self.lr_metrics_updated = False
        self.intermediate: Optional[MetricsOutputs] = None
        self.final_results: Optional[MetricsOutputs] = None
        self.checkpoint_scores: Dict[str, float] = {}

    @property
    def export_config(self) -> Dict[str, Any]:
        ddp_info = get_ddp_info()
        ddp_d = None if ddp_info is None else ddp_info._asdict()
        return {
            "state_config": self.state.config,
            "valid_portion": self.config.valid_portion,
            "mixed_precision": self.config.mixed_precision,
            "clip_norm": self.config.clip_norm,
            "metrics": (
                None
                if self.metrics is None
                else self.metrics.__identifier__
                if not isinstance(self.metrics, MultipleMetrics)
                else [metric.__identifier__ for metric in self.metrics.metrics]
            ),
            "loss_metrics_weights": self.config.loss_metrics_weights,
            "monitors": [monitor.__identifier__ for monitor in self.monitors],
            "callbacks": [callback.__identifier__ for callback in self.callbacks],
            "optimizer_settings": self.config.optimizer_settings,
            "ddp_info": ddp_d,
            "finetune_config": self.config.finetune_config,
            "tqdm_settings": self.tqdm_settings.asdict(),
        }

    @property
    def device(self) -> torch.device:
        return self.accelerator.device

    @property
    def is_local_rank_0(self) -> bool:
        return self.accelerator.is_local_main_process

    @property
    def use_tqdm_in_validation(self) -> bool:
        if not self.is_local_rank_0:
            return False
        if self.tqdm_settings.in_distributed:
            return False
        return self.tqdm_settings.use_tqdm_in_validation or self.state.is_terminate

    @property
    def validation_loader(self) -> IDataLoader:
        return self.valid_loader or self.train_loader_copy

    @property
    def input_sample(self) -> tensor_dict_type:
        return get_input_sample(self.train_loader_copy, self.device)

    @property
    def has_checkpoint_folder(self) -> bool:
        if self.checkpoint_folder is None:
            return False
        return os.path.isdir(self.checkpoint_folder)

    @property
    def workspace(self) -> str:
        return self.config.workspace

    @property
    def checkpoint_folder(self) -> str:
        return os.path.join(self.workspace, CHECKPOINTS_FOLDER)

    # init

    def _init_finetune(self) -> None:
        finetune_config = self.config.finetune_config
        if finetune_config is None:
            return None
        pretrained_ckpt = finetune_config.get("pretrained_ckpt")
        if pretrained_ckpt is None:
            raise ValueError("`rank` should be provided when `finetune` is triggered")
        print_info(f"loading pretrained checkpoint from '{pretrained_ckpt}'...")
        d = torch.load(pretrained_ckpt, map_location=self.device)
        self.model.load_state_dict(d)
        freeze = finetune_config.get("freeze", "")
        freeze_except = finetune_config.get("freeze_except", "")
        if not freeze and not freeze_except:
            return None
        msg_fmt = f"-> {'{}'} parameters will be {'{}'} under '{'{}'}'"
        param_names = []
        if freeze:
            num_frozen = 0
            for name, param in self.model.named_parameters():
                if re.match(freeze, name):
                    num_frozen += 1
                    param.requires_grad_(False)
                    param_names.append(name)
            msg = msg_fmt.format(num_frozen, "frozen", freeze)
        elif freeze_except:
            num_trainable = 0
            for name, param in self.model.named_parameters():
                if not re.match(freeze_except, name):
                    param.requires_grad_(False)
                else:
                    num_trainable += 1
                    param_names.append(name)
            msg = msg_fmt.format(num_trainable, "trainable", freeze_except)
        else:
            msg = "`freeze` & `freeze_except` should not be provided simultaneously"
            raise ValueError(msg)
        print("\n".join(["=" * 100, msg, "-" * 100] + param_names + ["-" * 100]))

    # core

    def post_loss_step(self, loss_dict: tensor_dict_type) -> None:
        # backward
        loss = loss_dict[LOSS_KEY]
        self.accelerator.backward(loss)
        # clip norm
        self.clip_norm_step()
        # optimize
        self.optimizer_step()
        self.scheduler_step()

    def weighted_loss_score(self, loss_items: Dict[str, float]) -> float:
        if not self.config.loss_metrics_weights:
            loss = loss_items.get(LOSS_KEY)
            if loss is not None:
                return -loss
            return -sum(loss_items.values()) / len(loss_items)
        score = 0.0
        for k, w in self.config.loss_metrics_weights.items():
            v = loss_items.get(k)
            if v is None:
                continue
            score -= v * w
        return score

    def clip_norm_step(self) -> None:
        if self.config.clip_norm > 0.0:
            if self.accelerator.sync_gradients:
                self._gradient_norm = self.accelerator.clip_grad_norm_(
                    self.model_for_training.parameters(),
                    max_norm=self.config.clip_norm,
                )

    def optimizer_step(self) -> None:
        for opt in self.optimizers.values():
            opt.step()
        for param in self.model_for_training.parameters():
            param.grad = None

    def scheduler_step(self) -> None:
        if self.config.update_scheduler_per_epoch:
            if self.state.epoch == self._current_scheduler_epoch:
                return
        lr_metric_logged = False
        for key, scheduler in self.schedulers.items():
            if scheduler is not None:
                should_log_lr, kwargs = self._get_scheduler_settings(key, scheduler)
                if should_log_lr or self.config.update_scheduler_per_epoch:
                    lr_metric_logged = True
                    for callback in self.callbacks:
                        callback.log_lr(
                            f"lr-{key}",
                            scheduler.get_last_lr()[0],
                            self.state,
                        )
                scheduler.step(**shallow_copy_dict(kwargs))
        if lr_metric_logged:
            self.lr_metrics_updated = False
        if self.config.update_scheduler_per_epoch:
            self._current_scheduler_epoch = self.state.epoch

    def _get_metrics(
        self,
        *,
        portion: float = 1.0,
        loader: Optional[IDataLoader] = None,
    ) -> MetricsOutputs:
        return get_metrics(
            self.config,
            self.model,
            self.loss,
            self.metrics,
            self.inference,
            loader or self.validation_loader,
            portion=portion,
            state=self.state,
        )

    def _get_scheduler_settings(
        self,
        key: str,
        scheduler: Any,
    ) -> Tuple[bool, Dict[str, Any]]:
        kwargs = {}
        should_log_lr = self.state.should_log_lr
        is_warmup = isinstance(scheduler, WarmupScheduler)
        requires_metric = key in self.schedulers_requires_metric
        if requires_metric and not (is_warmup and not scheduler.finished_warmup):
            if self.intermediate is None:
                kwargs["metrics"] = -math.inf
            else:
                kwargs["metrics"] = self.intermediate.final_score
            should_log_lr &= self.lr_metrics_updated
        return should_log_lr, kwargs

    def _logging_step(self, metrics_outputs: MetricsOutputs) -> None:
        if not self.is_local_rank_0:
            return None
        if self.epoch_tqdm is not None:
            metric_values = shallow_copy_dict(metrics_outputs.metric_values)
            metric_values["score"] = metrics_outputs.final_score
            self.epoch_tqdm.set_postfix(metric_values)
        for callback in self.callbacks:
            callback.log_metrics(metrics_outputs, self.state)
        if self.state.should_log_artifacts:
            for callback in self.callbacks:
                callback.log_artifacts(self)
        if self.state.should_log_metrics_msg:
            for callback in self.callbacks:
                callback.log_metrics_msg(
                    metrics_outputs,
                    self.metrics_log_path,
                    self.state,
                )

    def _monitor_step(self, step_outputs: StepOutputs) -> MonitorResults:
        terminate = False
        save_checkpoint = False
        for monitor in self.monitors:
            monitor.handle_extension(self.state)
        if self.state.should_monitor:
            # get metrics
            if (
                self.valid_loader is not None
                or self.config.recompute_train_losses_in_eval
            ):
                self.intermediate = self._get_metrics(portion=self.config.valid_portion)
            else:
                loss_dict = step_outputs.loss_dict
                is_positive = {k: False for k in loss_dict}
                loss_score = weighted_loss_score(self.config, loss_dict)
                self.intermediate = MetricsOutputs(loss_score, loss_dict, is_positive)
            self.lr_metrics_updated = True
            # logging
            self._logging_step(self.intermediate)
            # check terminate
            if self.state.should_start_snapshot:
                score = self.intermediate.final_score
                if any(monitor.snapshot(score) for monitor in self.monitors):
                    if self.state.can_snapshot:
                        self.state.update_snapshot_epoch()
                        save_checkpoint = True
                if any(monitor.check_terminate(score) for monitor in self.monitors):
                    terminate = True
        return MonitorResults(terminate, save_checkpoint, self.intermediate)

    def _step(self, batch_idx: int, batch: tensor_dict_type) -> StepOutputs:
        batch = to_device(batch, self.device)
        # kwargs
        forward_kw: Dict[str, Any] = {}
        for callback in self.callbacks:
            callback.mutate_train_forward_kwargs(forward_kw, self)
        loss_kw: Dict[str, Any] = {}
        for callback in self.callbacks:
            callback.mutate_train_loss_kwargs(loss_kw, self)
        # allow model defines its own training step
        if (
            isinstance(self.model, ModelWithCustomSteps)
            and self.model.custom_train_step
        ):
            return self.model.train_step(batch_idx, batch, self, forward_kw, loss_kw)
        # forward & loss
        forward_results = self.model.run(batch_idx, batch, self.state, **forward_kw)
        loss_dict = self.loss.run(forward_results, batch, self.state, **loss_kw)
        # post loss step
        self.post_loss_step(loss_dict)
        return StepOutputs(forward_results, {k: v.item() for k, v in loss_dict.items()})

    # api

    def fit(
        self,
        data: IData,
        loss: ILoss,
        model: IDLModel,
        metrics: Optional[IMetric],
        inference: IInference,
        optimizers: Dict[str, Optimizer],
        schedulers: Dict[str, Optional[_LRScheduler]],
        monitors: List[TrainerMonitor],
        callbacks: List[TrainerCallback],
        schedulers_requires_metric: Set[str],
        *,
        config_export_file: Optional[str] = None,
        show_summary: Optional[bool] = None,
        cuda: Optional[str] = None,
    ) -> "Trainer":
        # accelerator
        cpu = False
        if get_ddp_info() is None:
            if cuda is None:
                cpu = True
            else:
                os.environ["CUDA_VISIBLE_DEVICES"] = cuda
        if isinstance(self.config.mixed_precision, PrecisionType):
            self.config.mixed_precision = self.config.mixed_precision.value
        self.accelerator = Accelerator(
            cpu=cpu,
            mixed_precision=self.config.mixed_precision,
        )
        # initialize artifact structure
        if self.is_local_rank_0:
            os.makedirs(self.workspace, exist_ok=True)
            self.metrics_log_path = os.path.join(self.workspace, self.metrics_log_file)
            with open(self.metrics_log_path, "w"):
                pass
            os.makedirs(self.checkpoint_folder, exist_ok=True)
        # initialize
        self.model = model
        self.metrics = metrics
        self.monitors = monitors
        self.callbacks = callbacks
        self.schedulers_requires_metric = schedulers_requires_metric
        if self.is_local_rank_0:
            with open(os.path.join(self.workspace, self.model_log_file), "w") as f:
                f.write(str(model))
        self.inference = inference
        # data
        train_loader, valid_loader = data.get_loaders()
        self.train_loader = train_loader
        self.train_loader_copy = train_loader.copy()
        self.train_loader_copy.disable_shuffle()
        self.valid_loader = valid_loader
        if self.config.fixed_epoch is not None:
            num_epoch = max_epoch = self.config.fixed_epoch
        else:
            max_epoch = self.config.max_epoch
            num_epoch = min(max_epoch, self.config.num_epoch)
        self.state = TrainerState(
            train_loader,
            num_epoch=num_epoch,
            max_epoch=max_epoch,
            fixed_steps=self.config.fixed_steps,
            **(self.config.state_config or {}),
        )
        # accelerator prepare
        optimizer_keys = sorted(optimizers)
        (
            self.loss,
            self.model_for_training,
            *optimizer_list,
        ) = self.accelerator.prepare(
            loss,
            model,
            *[optimizers[k] for k in optimizer_keys],
        )
        self.optimizers = {k: optimizer_list[i] for i, k in enumerate(optimizer_keys)}
        self.schedulers = schedulers
        for sch in schedulers.values():
            if sch is not None:
                sch.load_state_dict(sch.state_dict())
        # callback
        self.model.init_with_trainer(self)
        # finetune
        self._init_finetune()
        # verbose
        if show_summary is None:
            show_summary = not self.tqdm_settings.in_distributed
        if self.is_local_rank_0:
            summary_msg = summary(
                self.model,
                to_device(self.input_sample, self.device),
                return_only=not show_summary,
            )
            with open(os.path.join(self.workspace, self.summary_log_file), "w") as f:
                f.write(summary_msg)
        # tqdm
        step_tqdm = None
        self.epoch_tqdm: Optional[tqdm] = None
        if self.is_local_rank_0 and self.tqdm_settings.use_tqdm:
            self.epoch_tqdm = tqdm(
                list(range(self.state.num_epoch)),
                position=self.tqdm_settings.position,
                desc=self.tqdm_settings.desc,
                leave=False,
            )
        # train
        has_ckpt = terminate = False
        if self.is_local_rank_0 and self.epoch_tqdm is None:
            print_info("entered training loop")
        if self.is_local_rank_0 and config_export_file is not None:
            config_export_path = os.path.join(self.workspace, config_export_file)
            with open(config_export_path, "w") as f:
                json.dump(self.export_config, f)
        for callback in self.callbacks:
            callback.before_loop(self)
        while self.state.should_train:
            try:
                self.state.epoch += 1
                if isinstance(self.train_loader, TorchDataLoader):
                    sampler = self.train_loader.loader.sampler
                    if isinstance(sampler, DistributedSampler):
                        sampler.set_epoch(self.state.epoch)
                        if isinstance(self.valid_loader, TorchDataLoader):
                            valid_sampler = self.valid_loader.loader.sampler
                            if isinstance(valid_sampler, DistributedSampler):
                                valid_sampler.set_epoch(self.state.epoch)
                step_iterator = TensorBatcher(self.train_loader, self.device)
                if self.is_local_rank_0 and self.tqdm_settings.use_step_tqdm:
                    step_tqdm = step_iterator = tqdm(
                        step_iterator,
                        total=len(self.train_loader),
                        position=self.tqdm_settings.position + 1,
                        leave=False,
                    )
                for i, batch in enumerate(step_iterator):
                    self.state.step += 1
                    step_outputs = self._step(i, batch)
                    for callback in self.callbacks:
                        callback.after_step(step_outputs, self.state)
                    monitor_results = self._monitor_step(step_outputs)
                    for callback in self.callbacks:
                        callback.after_monitor(monitor_results, self.state)
                    if self.is_local_rank_0 and monitor_results.save_checkpoint:
                        metric_outputs = monitor_results.metric_outputs
                        assert metric_outputs is not None
                        self.save_checkpoint(metric_outputs.final_score)
                    terminate = monitor_results.terminate or self.state.should_terminate
                    if terminate:
                        break
            except KeyboardInterrupt:
                if dist.is_initialized():
                    raise
                print_error("keyboard interrupted")
                terminate = True
            if terminate:
                break
            if self.epoch_tqdm is not None:
                self.epoch_tqdm.total = self.state.num_epoch
                self.epoch_tqdm.update()
        if self.epoch_tqdm is not None:
            if step_tqdm is not None:
                step_tqdm.close()
            self.epoch_tqdm.close()
        # restore
        if self.is_local_rank_0 and self.has_checkpoint_folder:
            if not self.tqdm_settings.in_distributed:
                print_info("rolling back to the best checkpoint")
            has_ckpt = self.restore_checkpoint()
        # finalize
        self.state.set_terminate()
        if self.is_local_rank_0:
            self.final_results = self._get_metrics(portion=self.config.valid_portion)
            self._logging_step(self.final_results)
            if not has_ckpt:
                self.save_checkpoint(self.final_results.final_score)
        for callback in self.callbacks:
            callback.finalize(self)
        return self

    # checkpointing

    def save_checkpoint(
        self,
        score: float,
        folder: Optional[str] = None,
        *,
        no_history: bool = False,
    ) -> None:
        if not self.is_local_rank_0:
            msg = "`save_checkpoint` should not be called when not `is_local_rank_0`"
            raise ValueError(msg)
        if folder is None:
            if self.checkpoint_folder is None:
                msg = "either `folder` or `checkpoint_folder` should be provided"
                raise ValueError(msg)
            folder = self.checkpoint_folder
        state = getattr(self, "state", None)
        pt_file = f"{PT_PREFIX}{-1 if state is None else state.step}.pt"
        if state is None:
            print_warning(
                "`state` is not initialized, "
                "latest model will be saved and the recorded score will always be 0"
            )
            torch.save(self.model.state_dict(), os.path.join(folder, pt_file))
            with open(os.path.join(folder, SCORES_FILE), "w") as f:
                json.dump({pt_file: 0.0}, f)
            return
        # leave top_k snapshots only
        if state.max_snapshot_file > 0:
            checkpoints = get_sorted_checkpoints(folder)
            if len(checkpoints) >= state.max_snapshot_file:
                for file in checkpoints[state.max_snapshot_file - 1 :]:
                    self.checkpoint_scores.pop(file)
                    os.remove(os.path.join(folder, file))
        # pt
        torch.save(self.model.state_dict(), os.path.join(folder, pt_file))
        # scores
        scores = {} if no_history else self.checkpoint_scores
        scores[pt_file] = score
        with open(os.path.join(folder, SCORES_FILE), "w") as f:
            json.dump(sort_dict_by_value(scores, reverse=True), f)

    def restore_checkpoint(
        self,
        folder: Optional[str] = None,
        strict: bool = True,
        state_dict_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
    ) -> bool:
        if not self.is_local_rank_0:
            msg = "`restore_checkpoint` should not be called when not `is_local_rank_0`"
            raise ValueError(msg)
        if folder is None:
            if self.checkpoint_folder is None:
                msg = "either `folder` or `checkpoint_folder` should be provided"
                raise ValueError(msg)
            folder = self.checkpoint_folder
        checkpoints = get_sorted_checkpoints(folder)
        if not checkpoints:
            if not self.tqdm_settings.in_distributed:
                print_warning(f"no model file found in {folder}")
            return False
        success = False
        for checkpoint in checkpoints:
            model_file = os.path.join(folder, checkpoint)
            if not os.path.isfile(model_file):
                continue
            if not self.tqdm_settings.in_distributed:
                print_info(f"restoring from {model_file}")
            states = torch.load(model_file, map_location=self.device)
            if state_dict_callback is not None:
                state_dict_callback(states)
            self.model.load_state_dict(states, strict)
            success = True
            break
        return success


__all__ = [
    "get_scores",
    "get_sorted_checkpoints",
    "Trainer",
]
